from collections import OrderedDict

from django.apps import apps
from django.conf import settings
from django.core.exceptions import FieldDoesNotExist
from django.http import Http404
from django.shortcuts import redirect
from django.urls import path, reverse
from django.utils.functional import classproperty
from modelcluster.fields import ParentalKey
from rest_framework import status
from rest_framework.renderers import BrowsableAPIRenderer, JSONRenderer
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet

from wagtail.api import APIField
from wagtail.models import Page, PageViewRestriction, Site

from .filters import (
    AncestorOfFilter,
    ChildOfFilter,
    DescendantOfFilter,
    FieldsFilter,
    LocaleFilter,
    OrderingFilter,
    SearchFilter,
    TranslationOfFilter,
)
from .pagination import WagtailPagination
from .serializers import BaseSerializer, PageSerializer, get_serializer_class
from .utils import (
    BadRequestError,
    get_object_detail_url,
    page_models_from_string,
    parse_fields_parameter,
)


class BaseAPIViewSet(GenericViewSet):
    @classproperty
    def renderer_classes(cls):
        renderers = [JSONRenderer]

        # Only add BrowsableAPIRenderer if rest_framework is installed
        # (which provides the necessary templates and static files)
        if apps.is_installed("rest_framework"):
            renderers.append(BrowsableAPIRenderer)

        return renderers

    pagination_class = WagtailPagination
    base_serializer_class = BaseSerializer
    filter_backends = []
    model = None  # Set on subclass

    known_query_parameters = frozenset(
        [
            "limit",
            "offset",
            "fields",
            "order",
            "search",
            "search_operator",
            # Used by jQuery for cache-busting. See #1671
            "_",
            # Required by BrowsableAPIRenderer
            "format",
        ]
    )
    find_query_parameters = frozenset(["id"])
    body_fields = ["id"]
    meta_fields = ["type", "detail_url"]
    listing_default_fields = ["id", "type", "detail_url"]
    nested_default_fields = ["id", "type", "detail_url"]
    detail_only_fields = []
    name = None  # Set on subclass.

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # seen_types is a mapping of type name strings (format: "app_label.ModelName")
        # to model classes. When an object is serialised in the API, its model
        # is added to this mapping. This is used by the Admin API which appends a
        # summary of the used types to the response.
        self.seen_types = OrderedDict()

    def get_queryset(self):
        return self.model.objects.all().order_by("id")

    def listing_view(self, request):
        queryset = self.get_queryset()
        self.check_query_parameters(queryset)
        queryset = self.filter_queryset(queryset)
        queryset = self.paginate_queryset(queryset)
        serializer = self.get_serializer(queryset, many=True)
        return self.get_paginated_response(serializer.data)

    def detail_view(self, request, pk):
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        return Response(serializer.data)

    def find_view(self, request):
        queryset = self.get_queryset()

        try:
            obj = self.find_object(queryset, request)

            if obj is None:
                raise self.model.DoesNotExist

        except self.model.DoesNotExist:
            raise Http404("not found")

        # Generate redirect
        url = get_object_detail_url(
            self.request.wagtailapi_router, request, self.model, obj.pk
        )

        if url is None:
            # Shouldn't happen unless this endpoint isn't actually installed in the router
            raise Exception(
                "Cannot generate URL to detail view. Is '{}' installed in the API router?".format(
                    self.__class__.__name__
                )
            )

        # Retain all query parameters except ones only used to find the object
        query = request.GET.copy()
        for param in self.find_query_parameters:
            query.pop(param, None)

        return redirect(f"{url}?{query.urlencode()}")

    def find_object(self, queryset, request):
        """
        Override this to implement more find methods.
        """
        if "id" in request.GET:
            return queryset.get(id=request.GET["id"])

    def handle_exception(self, exc):
        if isinstance(exc, Http404):
            data = {"message": str(exc)}
            return Response(data, status=status.HTTP_404_NOT_FOUND)
        elif isinstance(exc, BadRequestError):
            data = {"message": str(exc)}
            return Response(data, status=status.HTTP_400_BAD_REQUEST)
        return super().handle_exception(exc)

    @classmethod
    def _convert_api_fields(cls, fields):
        return [
            field if isinstance(field, APIField) else APIField(field)
            for field in fields
        ]

    @classmethod
    def get_body_fields(cls, model):
        return cls._convert_api_fields(
            cls.body_fields + list(getattr(model, "api_fields", ()))
        )

    @classmethod
    def get_body_fields_names(cls, model):
        return [field.name for field in cls.get_body_fields(model)]

    @classmethod
    def get_meta_fields(cls, model):
        return cls._convert_api_fields(
            cls.meta_fields + list(getattr(model, "api_meta_fields", ()))
        )

    @classmethod
    def get_meta_fields_names(cls, model):
        return [field.name for field in cls.get_meta_fields(model)]

    @classmethod
    def get_field_serializer_overrides(cls, model):
        return {
            field.name: field.serializer
            for field in cls.get_body_fields(model) + cls.get_meta_fields(model)
            if field.serializer is not None
        }

    @classmethod
    def get_available_fields(cls, model, db_fields_only=False):
        """
        Returns a list of all the fields that can be used in the API for the
        specified model class.

        Setting db_fields_only to True will remove all fields that do not have
        an underlying column in the database (eg, type/detail_url and any custom
        fields that are callables)
        """
        fields = cls.get_body_fields_names(model) + cls.get_meta_fields_names(model)

        if db_fields_only:
            # Get list of available database fields then remove any fields in our
            # list that isn't a database field
            database_fields = set()
            for field in model._meta.get_fields():
                database_fields.add(field.name)

                if hasattr(field, "attname"):
                    database_fields.add(field.attname)

            fields = [field for field in fields if field in database_fields]

        return fields

    @classmethod
    def get_detail_default_fields(cls, model):
        return cls.get_available_fields(model)

    @classmethod
    def get_listing_default_fields(cls, model):
        return cls.listing_default_fields[:]

    @classmethod
    def get_nested_default_fields(cls, model):
        return cls.nested_default_fields[:]

    def check_query_parameters(self, queryset):
        """
        Ensure that only valid query parameters are included in the URL.
        """
        query_parameters = set(self.request.GET.keys())

        # All query parameters must be either a database field or an operation
        allowed_query_parameters = set(
            self.get_available_fields(queryset.model, db_fields_only=True)
        ).union(self.known_query_parameters)
        unknown_parameters = query_parameters - allowed_query_parameters
        if unknown_parameters:
            raise BadRequestError(
                "query parameter is not an operation or a recognised field: %s"
                % ", ".join(sorted(unknown_parameters))
            )

    @classmethod
    def _get_serializer_class(
        cls, router, model, fields_config, show_details=False, nested=False
    ):
        # Get all available fields
        body_fields = cls.get_body_fields_names(model)
        meta_fields = cls.get_meta_fields_names(model)
        all_fields = body_fields + meta_fields

        # Remove any duplicates
        all_fields = list(OrderedDict.fromkeys(all_fields))

        if not show_details:
            # Remove detail only fields
            for field in cls.detail_only_fields:
                try:
                    all_fields.remove(field)
                except ValueError:
                    pass

        # Get list of configured fields
        if show_details:
            fields = set(cls.get_detail_default_fields(model))
        elif nested:
            fields = set(cls.get_nested_default_fields(model))
        else:
            fields = set(cls.get_listing_default_fields(model))

        # If first field is '*' start with all fields
        # If first field is '_' start with no fields
        if fields_config and fields_config[0][0] == "*":
            fields = set(all_fields)
            fields_config = fields_config[1:]
        elif fields_config and fields_config[0][0] == "_":
            fields = set()
            fields_config = fields_config[1:]

        mentioned_fields = set()
        sub_fields = {}

        for field_name, negated, field_sub_fields in fields_config:
            if negated:
                try:
                    fields.remove(field_name)
                except KeyError:
                    pass
            else:
                fields.add(field_name)
                if field_sub_fields:
                    sub_fields[field_name] = field_sub_fields

            mentioned_fields.add(field_name)

        unknown_fields = mentioned_fields - set(all_fields)

        if unknown_fields:
            raise BadRequestError(
                "unknown fields: %s" % ", ".join(sorted(unknown_fields))
            )

        # Build nested serialisers
        child_serializer_classes = {}

        for field_name in fields:
            try:
                django_field = model._meta.get_field(field_name)
            except FieldDoesNotExist:
                django_field = None

            if django_field and django_field.is_relation:
                child_sub_fields = sub_fields.get(field_name, [])

                # Inline (aka "child") models should display all fields by default
                if isinstance(getattr(django_field, "field", None), ParentalKey):
                    if not child_sub_fields or child_sub_fields[0][0] not in ["*", "_"]:
                        child_sub_fields = list(child_sub_fields)
                        child_sub_fields.insert(0, ("*", False, None))

                # Get a serializer class for the related object
                child_model = django_field.related_model
                child_endpoint_class = router.get_model_endpoint(child_model)
                child_endpoint_class = (
                    child_endpoint_class[1] if child_endpoint_class else BaseAPIViewSet
                )
                child_serializer_classes[field_name] = (
                    child_endpoint_class._get_serializer_class(
                        router, child_model, child_sub_fields, nested=True
                    )
                )

            else:
                if field_name in sub_fields:
                    # Sub fields were given for a non-related field
                    raise BadRequestError(
                        "'%s' does not support nested fields" % field_name
                    )

        # Reorder fields so it matches the order of all_fields
        fields = [field for field in all_fields if field in fields]

        field_serializer_overrides = {
            field[0]: field[1]
            for field in cls.get_field_serializer_overrides(model).items()
            if field[0] in fields
        }
        return get_serializer_class(
            model,
            fields,
            meta_fields=meta_fields,
            field_serializer_overrides=field_serializer_overrides,
            child_serializer_classes=child_serializer_classes,
            base=cls.base_serializer_class,
        )

    def get_serializer_class(self):
        request = self.request

        # Get model
        if self.action == "listing_view":
            model = self.get_queryset().model
        else:
            model = type(self.get_object())

        # Fields
        if "fields" in request.GET:
            try:
                fields_config = parse_fields_parameter(request.GET["fields"])
            except ValueError as e:
                raise BadRequestError("fields error: %s" % str(e))
        else:
            # Use default fields
            fields_config = []

        # Allow "detail_only" (eg parent) fields on detail view
        if self.action == "listing_view":
            show_details = False
        else:
            show_details = True

        return self._get_serializer_class(
            self.request.wagtailapi_router,
            model,
            fields_config,
            show_details=show_details,
        )

    def get_serializer_context(self):
        """
        The serialization context differs between listing and detail views.
        """
        return {
            "request": self.request,
            "view": self,
            "router": self.request.wagtailapi_router,
        }

    def get_renderer_context(self):
        context = super().get_renderer_context()
        context["indent"] = 4
        return context

    @classmethod
    def get_urlpatterns(cls):
        """
        This returns a list of URL patterns for the endpoint
        """
        return [
            path("", cls.as_view({"get": "listing_view"}), name="listing"),
            path("<int:pk>/", cls.as_view({"get": "detail_view"}), name="detail"),
            path("find/", cls.as_view({"get": "find_view"}), name="find"),
        ]

    @classmethod
    def get_model_listing_urlpath(cls, model, namespace=""):
        if namespace:
            url_name = namespace + ":listing"
        else:
            url_name = "listing"

        return reverse(url_name)

    @classmethod
    def get_object_detail_urlpath(cls, model, pk, namespace=""):
        if namespace:
            url_name = namespace + ":detail"
        else:
            url_name = "detail"

        return reverse(url_name, args=(pk,))


class PagesAPIViewSet(BaseAPIViewSet):
    base_serializer_class = PageSerializer
    filter_backends = [
        FieldsFilter,
        ChildOfFilter,
        AncestorOfFilter,
        DescendantOfFilter,
        OrderingFilter,
        TranslationOfFilter,
        LocaleFilter,
        SearchFilter,  # needs to be last, as SearchResults querysets cannot be filtered further
    ]
    known_query_parameters = BaseAPIViewSet.known_query_parameters.union(
        [
            "type",
            "child_of",
            "ancestor_of",
            "descendant_of",
            "translation_of",
            "locale",
            "site",
        ]
    )
    find_query_parameters = BaseAPIViewSet.find_query_parameters.union(
        [
            "html_path",
        ]
    )
    body_fields = BaseAPIViewSet.body_fields + [
        "title",
    ]
    meta_fields = BaseAPIViewSet.meta_fields + [
        "html_url",
        "slug",
        "show_in_menus",
        "seo_title",
        "search_description",
        "first_published_at",
        "alias_of",
        "parent",
        "locale",
    ]
    listing_default_fields = BaseAPIViewSet.listing_default_fields + [
        "title",
        "html_url",
        "slug",
        "first_published_at",
    ]
    nested_default_fields = BaseAPIViewSet.nested_default_fields + [
        "title",
    ]
    detail_only_fields = ["parent"]
    name = "pages"
    model = Page

    @classmethod
    def get_detail_default_fields(cls, model):
        detail_default_fields = super().get_detail_default_fields(model)

        # When i18n is disabled, remove "locale" from default fields
        if not getattr(settings, "WAGTAIL_I18N_ENABLED", False):
            detail_default_fields.remove("locale")

        return detail_default_fields

    @classmethod
    def get_listing_default_fields(cls, model):
        listing_default_fields = super().get_listing_default_fields(model)

        # When i18n is enabled, add "locale" to default fields
        if getattr(settings, "WAGTAIL_I18N_ENABLED", False):
            listing_default_fields.append("locale")

        return listing_default_fields

    def get_root_page(self):
        """
        Returns the page that is used when the `&child_of=root` filter is used.
        """
        return Site.find_for_request(self.request).root_page

    def get_base_queryset(self):
        """
        Returns a queryset containing all pages that can be seen by this user.

        This is used as the base for get_queryset and is also used to find the
        parent pages when using the child_of and descendant_of filters as well.
        """

        request = self.request

        # Get all live pages
        queryset = Page.objects.all().live()

        # Exclude pages that the user doesn't have access to
        restricted_pages = [
            restriction.page
            for restriction in PageViewRestriction.objects.all().select_related("page")
            if not restriction.accept_request(self.request)
        ]

        # Exclude the restricted pages and their descendants from the queryset
        for restricted_page in restricted_pages:
            queryset = queryset.not_descendant_of(restricted_page, inclusive=True)

        # Check if we have a specific site to look for
        if "site" in request.GET:
            # Optionally allow querying by port
            if ":" in request.GET["site"]:
                (hostname, port) = request.GET["site"].split(":", 1)
                query = {
                    "hostname": hostname,
                    "port": port,
                }
            else:
                query = {
                    "hostname": request.GET["site"],
                }
            try:
                site = Site.objects.get(**query)
            except Site.MultipleObjectsReturned:
                raise BadRequestError(
                    "Your query returned multiple sites. Try adding a port number to your site filter."
                )
        else:
            # Otherwise, find the site from the request
            site = Site.find_for_request(self.request)

        if site:
            base_queryset = queryset
            queryset = base_queryset.descendant_of(site.root_page, inclusive=True)

            # If internationalisation is enabled, include pages from other language trees
            if getattr(settings, "WAGTAIL_I18N_ENABLED", False):
                for translation in site.root_page.get_translations():
                    queryset |= base_queryset.descendant_of(translation, inclusive=True)

        else:
            # No sites configured
            queryset = queryset.none()

        return queryset

    def get_queryset(self):
        request = self.request

        # Allow pages to be filtered to a specific type
        try:
            models_type = request.GET.get("type", None)
            models = models_type and page_models_from_string(models_type) or []
        except (LookupError, ValueError):
            raise BadRequestError("type doesn't exist")

        if not models:
            if self.model == Page:
                return self.get_base_queryset()
            else:
                return self.model.objects.filter(
                    pk__in=self.get_base_queryset().values_list("pk", flat=True)
                )

        elif len(models) == 1:
            # If a single page type has been specified, swap out the Page-based queryset for one based on
            # the specific page model so that we can filter on any custom APIFields defined on that model
            return models[0].objects.filter(
                pk__in=self.get_base_queryset().values_list("pk", flat=True)
            )

        else:  # len(models) > 1
            return self.get_base_queryset().type(*models)

    def get_object(self):
        base = super().get_object()
        return base.specific

    def find_object(self, queryset, request):
        site = Site.find_for_request(request)
        if "html_path" in request.GET and site is not None:
            path = request.GET["html_path"]
            path_components = [component for component in path.split("/") if component]

            try:
                page, _, _ = site.root_page.specific.route(request, path_components)
            except Http404:
                return

            if queryset.filter(id=page.id).exists():
                return page

        return super().find_object(queryset, request)

    def get_serializer_context(self):
        """
        The serialization context differs between listing and detail views.
        """
        context = super().get_serializer_context()
        context["base_queryset"] = self.get_base_queryset()
        return context



import warnings

from wagtail.signals import init_new_page  # noqa: F401
from wagtail.utils.deprecation import RemovedInWagtail80Warning

warnings.warn(
    "wagtail.admin.signals.init_new_page has been moved to wagtail.signals.init_new_page",
    RemovedInWagtail80Warning,
    stacklevel=2,
)



"""Handles rendering of the list of actions in the footer of the page create/edit views."""

from django.conf import settings
from django.forms import Media
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from wagtail import hooks
from wagtail.admin.ui.components import Component


class ActionMenuItem(Component):
    """Defines an item in the actions drop-up on the page creation/edit view"""

    order = 100  # default order index if one is not specified on init
    template_name = "wagtailadmin/pages/action_menu/menu_item.html"

    label = ""
    name = None
    classname = ""
    icon_name = ""

    def __init__(self, order=None):
        if order is not None:
            self.order = order

    def get_user_page_permissions_tester(self, context):
        if "user_page_permissions_tester" in context:
            return context["user_page_permissions_tester"]
        return context["page"].permissions_for_user(context["request"].user)

    def is_shown(self, context):
        """
        Whether this action should be shown on this request; permission checks etc should go here.
        By default, actions are shown for unlocked pages, hidden for locked pages

        context = dictionary containing at least:
            'request' = the current request object
            'view' = 'create', 'edit' or 'revisions_revert'
            'page' (if view = 'edit' or 'revisions_revert') = the page being edited
            'parent_page' (if view = 'create') = the parent page of the page being created
            'lock' = a Lock object if the page is locked, otherwise None
            'locked_for_user' = True if the lock prevents the current user from editing the page
            may also contain:
            'user_page_permissions_tester' = a PagePermissionTester for the current user and page
        """
        return context["view"] == "create" or not context["locked_for_user"]

    def get_context_data(self, parent_context):
        """Defines context for the template, overridable to use more data"""
        context = parent_context.copy()
        url = self.get_url(parent_context)

        context.update(
            {
                "label": self.label,
                "url": url,
                "name": self.name,
                "classname": self.classname,
                "icon_name": self.icon_name,
                "request": parent_context["request"],
            }
        )
        return context

    def get_url(self, parent_context):
        return None


class PublishMenuItem(ActionMenuItem):
    label = _("Publish")
    name = "action-publish"
    template_name = "wagtailadmin/pages/action_menu/publish.html"
    icon_name = "upload"

    def is_shown(self, context):
        if context["view"] == "create":
            return (
                context["parent_page"]
                .permissions_for_user(context["request"].user)
                .can_publish_subpage()
            )
        else:  # view == 'edit' or 'revisions_revert'
            perms_tester = self.get_user_page_permissions_tester(context)
            return not context["locked_for_user"] and perms_tester.can_publish()

    def get_context_data(self, parent_context):
        context = super().get_context_data(parent_context)
        page = context.get("page")
        context["is_scheduled"] = page and page.go_live_at
        context["is_revision"] = context["view"] == "revisions_revert"
        return context


class SubmitForModerationMenuItem(ActionMenuItem):
    label = _("Submit for moderation")
    name = "action-submit"
    icon_name = "resubmit"

    def is_shown(self, context):
        if not getattr(settings, "WAGTAIL_WORKFLOW_ENABLED", True):
            return False

        if context["view"] == "create":
            return context["parent_page"].has_workflow

        if context["view"] == "edit":
            perms_tester = self.get_user_page_permissions_tester(context)
            return (
                perms_tester.can_submit_for_moderation()
                and not context["locked_for_user"]
            )
        # context == revisions_revert
        return False

    def get_context_data(self, parent_context):
        context = super().get_context_data(parent_context)
        page = context.get("page")
        workflow_state = page.current_workflow_state if page else None
        if (
            workflow_state
            and workflow_state.status == workflow_state.STATUS_NEEDS_CHANGES
        ):
            context["label"] = _("Resubmit to %(task_name)s") % {
                "task_name": workflow_state.current_task_state.task.name
            }
        elif page:
            workflow = page.get_workflow()
            if workflow:
                context["label"] = _("Submit to %(workflow_name)s") % {
                    "workflow_name": workflow.name
                }
        return context


class WorkflowMenuItem(ActionMenuItem):
    template_name = "wagtailadmin/pages/action_menu/workflow_menu_item.html"

    def __init__(self, name, label, launch_modal, *args, **kwargs):
        self.name = name
        self.label = label
        self.launch_modal = launch_modal

        if kwargs.get("icon_name"):
            self.icon_name = kwargs.pop("icon_name")

        super().__init__(*args, **kwargs)

    def get_context_data(self, parent_context):
        context = super().get_context_data(parent_context)
        context["launch_modal"] = self.launch_modal
        context["current_task_state"] = context["page"].current_workflow_task_state
        return context

    def is_shown(self, context):
        if context["view"] == "edit":
            return not context["locked_for_user"]


class RestartWorkflowMenuItem(ActionMenuItem):
    label = _("Restart workflow ")
    name = "action-restart-workflow"
    classname = "button--icon-flipped"
    icon_name = "login"

    def is_shown(self, context):
        if not getattr(settings, "WAGTAIL_WORKFLOW_ENABLED", True):
            return False
        elif context["view"] == "edit":
            workflow_state = context["page"].current_workflow_state
            perms_tester = self.get_user_page_permissions_tester(context)
            return (
                perms_tester.can_submit_for_moderation()
                and not context["locked_for_user"]
                and workflow_state
                and workflow_state.user_can_cancel(context["request"].user)
            )
        else:
            return False


class CancelWorkflowMenuItem(ActionMenuItem):
    label = _("Cancel workflow ")
    name = "action-cancel-workflow"
    icon_name = "error"

    def is_shown(self, context):
        if context["view"] == "edit":
            workflow_state = context["page"].current_workflow_state
            return workflow_state and workflow_state.user_can_cancel(
                context["request"].user
            )
        return False


class UnpublishMenuItem(ActionMenuItem):
    label = _("Unpublish")
    name = "action-unpublish"
    icon_name = "download"

    def is_shown(self, context):
        if context["view"] == "edit":
            perms_tester = self.get_user_page_permissions_tester(context)
            return not context["locked_for_user"] and perms_tester.can_unpublish()

    def get_url(self, context):
        return reverse("wagtailadmin_pages:unpublish", args=(context["page"].id,))


class SaveDraftMenuItem(ActionMenuItem):
    name = "action-save-draft"
    label = _("Save Draft")
    template_name = "wagtailadmin/pages/action_menu/save_draft.html"

    def get_context_data(self, parent_context):
        context = super().get_context_data(parent_context)
        context["is_revision"] = context["view"] == "revisions_revert"
        return context


class PageLockedMenuItem(ActionMenuItem):
    name = "action-page-locked"
    label = _("Page locked")
    template_name = "wagtailadmin/pages/action_menu/page_locked.html"

    def is_shown(self, context):
        return "page" in context and context["locked_for_user"]

    def get_context_data(self, parent_context):
        context = super().get_context_data(parent_context)
        context["is_revision"] = context["view"] == "revisions_revert"
        return context


BASE_PAGE_ACTION_MENU_ITEMS = None


def _get_base_page_action_menu_items():
    """
    Retrieve the global list of menu items for the page action menu,
    which may then be customized on a per-request basis
    """
    global BASE_PAGE_ACTION_MENU_ITEMS

    if BASE_PAGE_ACTION_MENU_ITEMS is None:
        BASE_PAGE_ACTION_MENU_ITEMS = [
            SaveDraftMenuItem(order=0),
            UnpublishMenuItem(order=20),
            PublishMenuItem(order=30),
            CancelWorkflowMenuItem(order=40),
            RestartWorkflowMenuItem(order=50),
            SubmitForModerationMenuItem(order=60),
            PageLockedMenuItem(order=10000),
        ]
        for hook in hooks.get_hooks("register_page_action_menu_item"):
            action_menu_item = hook()
            if action_menu_item:
                BASE_PAGE_ACTION_MENU_ITEMS.append(action_menu_item)

    return BASE_PAGE_ACTION_MENU_ITEMS


class PageActionMenu:
    template = "wagtailadmin/pages/action_menu/menu.html"

    def __init__(self, request, **kwargs):
        self.request = request
        self.context = kwargs
        self.context["request"] = request
        page = self.context.get("page")
        if page:
            self.context["user_page_permissions_tester"] = page.permissions_for_user(
                self.request.user
            )

        self.menu_items = []

        if page:
            task = page.current_workflow_task
            current_workflow_state = page.current_workflow_state
            is_final_task = (
                current_workflow_state and current_workflow_state.is_at_final_task
            )
            if task:
                actions = task.get_actions(page, request.user)
                workflow_menu_items = []
                for name, label, launch_modal in actions:
                    icon_name = "edit"
                    if name == "approve":
                        if is_final_task and not getattr(
                            settings,
                            "WAGTAIL_WORKFLOW_REQUIRE_REAPPROVAL_ON_EDIT",
                            False,
                        ):
                            label = _("%(label)s and Publish") % {"label": label}
                        icon_name = "success"

                    item = WorkflowMenuItem(
                        name, label, launch_modal, icon_name=icon_name
                    )

                    if item.is_shown(self.context):
                        workflow_menu_items.append(item)
                self.menu_items.extend(workflow_menu_items)

        for menu_item in _get_base_page_action_menu_items():
            if menu_item.is_shown(self.context):
                self.menu_items.append(menu_item)

        self.menu_items.sort(key=lambda item: item.order)

        for hook in hooks.get_hooks("construct_page_action_menu"):
            hook(self.menu_items, self.request, self.context)

        try:
            self.default_item = self.menu_items.pop(0)
        except IndexError:
            self.default_item = None

    def render_html(self):
        if not self.default_item:
            return ""

        rendered_menu_items = [
            menu_item.render_html(self.context) for menu_item in self.menu_items
        ]

        rendered_default_item = self.default_item.render_html(self.context)

        return render_to_string(
            self.template,
            {
                "default_menu_item": rendered_default_item,
                "show_menu": bool(self.menu_items),
                "rendered_menu_items": rendered_menu_items,
            },
            request=self.request,
        )

    @cached_property
    def media(self):
        media = self.default_item.media if self.default_item else Media()
        for item in self.menu_items:
            media += item.media
        return media



from wagtail.api.v2.filters import FieldsFilter, OrderingFilter, SearchFilter
from wagtail.api.v2.views import BaseAPIViewSet

from ... import get_image_model
from .serializers import ImageSerializer


class ImagesAPIViewSet(BaseAPIViewSet):
    base_serializer_class = ImageSerializer
    filter_backends = [FieldsFilter, OrderingFilter, SearchFilter]
    body_fields = BaseAPIViewSet.body_fields + ["title", "width", "height"]
    meta_fields = BaseAPIViewSet.meta_fields + ["tags", "download_url"]
    listing_default_fields = BaseAPIViewSet.listing_default_fields + [
        "title",
        "tags",
        "download_url",
    ]
    nested_default_fields = BaseAPIViewSet.nested_default_fields + [
        "title",
        "download_url",
    ]
    name = "images"
    model = get_image_model()



from django.utils import timezone
from django.utils.text import capfirst
from django.utils.translation import gettext as _

from wagtail.admin.utils import get_latest_str
from wagtail.admin.views.generic.base import BaseOperationView
from wagtail.log_actions import log


class LockView(BaseOperationView):
    success_message_extra_tags = "lock"

    def perform_operation(self):
        if self.object.locked:
            return
        self.object.locked = True
        self.object.locked_by = self.request.user
        self.object.locked_at = timezone.now()
        self.object.save(update_fields=["locked", "locked_by", "locked_at"])
        log(instance=self.object, action="wagtail.lock", user=self.request.user)


class UnlockView(BaseOperationView):
    success_message_extra_tags = "unlock"

    def perform_operation(self):
        if not self.object.locked:
            return
        self.object.locked = False
        self.object.locked_by = None
        self.object.locked_at = None
        self.object.save(update_fields=["locked", "locked_by", "locked_at"])
        log(instance=self.object, action="wagtail.unlock", user=self.request.user)

    def get_success_message(self):
        return capfirst(
            _("%(model_name)s '%(title)s' is now unlocked.")
            % {
                "model_name": self.model._meta.verbose_name,
                "title": get_latest_str(self.object),
            }
        )



from django.core.checks import Error
from django.test import TestCase, override_settings
from django.utils.formats import reset_format_cache

from wagtail.admin.checks import datetime_format_check
from wagtail.test.utils import WagtailTestUtils


class TestDateTimeChecks(WagtailTestUtils, TestCase):
    fixtures = ["test.json"]

    def setUp(self):
        reset_format_cache()

    def test_datetime_format(self):
        with override_settings(
            WAGTAIL_CONTENT_LANGUAGES=[
                ("en", "English"),
            ],
            LANGUAGES=[
                ("en", "English"),
            ],
            WAGTAIL_DATE_FORMAT="%m/%d/%Y",
            WAGTAIL_TIME_FORMAT="%H:%M",
            USE_L10N=True,
        ):
            errors = datetime_format_check(None)

        self.assertEqual(errors, [])

    def test_datetime_format_with_unsupported_date(self):
        with override_settings(
            WAGTAIL_CONTENT_LANGUAGES=[
                ("en", "English"),
            ],
            LANGUAGES=[
                ("en", "English"),
            ],
            WAGTAIL_DATE_FORMAT="%d.%m.%Y.",
            WAGTAIL_TIME_FORMAT="%H:%M",
            USE_L10N=True,
        ):
            errors = datetime_format_check(None)

        expected_errors = [
            Error(
                "Configuration error",
                hint="'%d.%m.%Y.' must be in DATE_INPUT_FORMATS for language English (en).",
                obj="WAGTAIL_DATE_FORMAT",
                id="wagtailadmin.E003",
            )
        ]
        self.assertEqual(errors, expected_errors)

    def test_datetime_format_with_unsupported_date_not_using_l10n(self):
        """
        Test that the check doesn't raise an error when USE_L10N is False.
        """

        with override_settings(
            WAGTAIL_CONTENT_LANGUAGES=[
                ("en", "English"),
            ],
            LANGUAGES=[
                ("en", "English"),
            ],
            WAGTAIL_DATE_FORMAT="%d.%m.%Y.",
            WAGTAIL_TIME_FORMAT="%H:%M",
            USE_L10N=False,
        ):
            errors = datetime_format_check(None)
        self.assertEqual(errors, [])

    def test_datetime_format_with_unsupported_datetime_and_time(self):
        with override_settings(
            WAGTAIL_CONTENT_LANGUAGES=[
                ("en", "English"),
            ],
            LANGUAGES=[
                ("en", "English"),
            ],
            WAGTAIL_DATETIME_FORMAT="%d.%m.%Y. %H:%M",
            WAGTAIL_TIME_FORMAT="%I:%M %p",
            USE_L10N=True,
        ):
            errors = datetime_format_check(None)

        expected_errors = [
            Error(
                "Configuration error",
                hint="'%d.%m.%Y. %H:%M' must be in DATETIME_INPUT_FORMATS for language English (en).",
                obj="WAGTAIL_DATETIME_FORMAT",
                id="wagtailadmin.E003",
            ),
            Error(
                "Configuration error",
                hint="'%I:%M %p' must be in TIME_INPUT_FORMATS for language English (en).",
                obj="WAGTAIL_TIME_FORMAT",
                id="wagtailadmin.E003",
            ),
        ]
        self.assertEqual(errors, expected_errors)

    def test_datetime_format_with_overriden_format(self):
        with override_settings(
            WAGTAIL_CONTENT_LANGUAGES=[
                ("en", "English"),
            ],
            LANGUAGES=[
                ("en", "English"),
            ],
            WAGTAIL_DATETIME_FORMAT="%d.%m.%Y. %H:%M",
            FORMAT_MODULE_PATH=["wagtail.admin.tests.formats"],
            USE_L10N=True,
        ):
            errors = datetime_format_check(None)

        self.assertEqual(errors, [])

    def test_datetime_format_with_incorrect_overriden_format(self):
        with override_settings(
            WAGTAIL_CONTENT_LANGUAGES=[
                ("en", "English"),
            ],
            LANGUAGES=[
                ("en", "English"),
            ],
            WAGTAIL_DATETIME_FORMAT="%m.%d.%Y. %H:%M",
            FORMAT_MODULE_PATH=["wagtail.admin.tests.formats"],
            USE_L10N=True,
        ):
            errors = datetime_format_check(None)

        expected_errors = [
            Error(
                "Configuration error",
                hint="'%m.%d.%Y. %H:%M' must be in DATETIME_INPUT_FORMATS for language English (en).",
                obj="WAGTAIL_DATETIME_FORMAT",
                id="wagtailadmin.E003",
            ),
        ]
        self.assertEqual(errors, expected_errors)



# Generated by Django 5.1.5 on 2025-02-13 01:36

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("tests", "0046_alter_custom_rendition_to_unique_constraint"),
        ("wagtailcore", "0094_alter_page_locale"),
    ]

    operations = [
        migrations.AddField(
            model_name="advertwithcustomuuidprimarykey",
            name="page",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="wagtailcore.page",
            ),
        ),
    ]



from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from wagtail import hooks
from wagtail.admin.menu import MenuItem
from wagtail.permissions import site_permission_policy

from .views import LocaleViewSet


@hooks.register("register_admin_viewset")
def register_viewset():
    return LocaleViewSet("wagtaillocales", url_prefix="locales")


class LocalesMenuItem(MenuItem):
    def is_shown(self, request):
        return site_permission_policy.user_has_any_permission(
            request.user, ["add", "change", "delete"]
        )


@hooks.register("register_settings_menu_item")
def register_locales_menu_item():
    return LocalesMenuItem(
        _("Locales"),
        reverse("wagtaillocales:index"),
        name="locales",
        icon_name="site",
        order=603,
    )



# This file contains a file storage backend that imitates behaviours of
# common external storage backends (S3 boto, libcloud, etc).

# The following behaviours have been added to this backend:
#  - Calling .path on the storage or image file raises NotImplementedError
#  - File.open() after the file has been closed raises an error
#  - File.size exceptions raise DummyExternalStorageError
#  - Storage._save() fails loudly if the content file's pointer is not at the start

from django.core.files import File
from django.core.files.storage import FileSystemStorage, Storage
from django.utils.deconstruct import deconstructible


@deconstructible
class DummyExternalStorage(Storage):
    def __init__(self, *args, **kwargs):
        self.wrapped = FileSystemStorage(*args, **kwargs)

    def path(self, name):
        # Overridden to give it the behaviour of the base Storage class
        # This is what an external storage backend would have
        raise NotImplementedError("This backend doesn't support absolute paths.")

    def _open(self, name, mode="rb"):
        # Overridden to return a DummyExternalStorageFile instead of a normal
        # File object
        return DummyExternalStorageFile(open(self.wrapped.path(name), mode))

    def _save(self, name, content):
        file_pos = content.tell()
        if file_pos != 0:
            raise ValueError(
                "Content file pointer should be at 0 - got %d instead" % file_pos
            )
        return self.wrapped._save(name, content)

    # Wrap all other functions

    def delete(self, name):
        self.wrapped.delete(name)

    def exists(self, name):
        return self.wrapped.exists(name)

    def listdir(self, path):
        return self.wrapped.listdir(path)

    def size(self, name):
        return self.wrapped.size(name)

    def url(self, name):
        return self.wrapped.url(name)

    def accessed_time(self, name):
        return self.wrapped.accessed_time(name)

    def created_time(self, name):
        return self.wrapped.created_time(name)

    def modified_time(self, name):
        return self.wrapped.modified_time(name)


class DummyExternalStorageError(Exception):
    pass


class DummyExternalStorageFile(File):
    def open(self, mode=None):
        # Based on:
        # https://github.com/django/django/blob/2c39f282b8389f47fee4b24e785a58567c6c3629/django/core/files/base.py#L135-L141

        # I've commented out two lines of this function which stops it checking
        # the filesystem for the file. Making it behave as if it is using an
        # external file storage.

        if not self.closed:
            self.seek(0)
        # elif self.name and os.path.exists(self.name):
        #    self.file = open(self.name, mode or self.mode)
        else:
            raise ValueError("The file cannot be reopened.")

    def size(self):
        try:
            return super().size
        except Exception as e:  # noqa: BLE001
            raise DummyExternalStorageError(str(e))



import datetime
import hashlib
import os
import random
import string
import uuid

from django import forms
from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.core.paginator import EmptyPage, PageNotAnInteger, Paginator
from django.db import models
from django.shortcuts import redirect
from django.template.response import TemplateResponse
from django.utils.translation import gettext_lazy as _
from modelcluster.contrib.taggit import ClusterTaggableManager
from modelcluster.fields import ParentalKey, ParentalManyToManyField
from modelcluster.models import ClusterableModel
from taggit.managers import TaggableManager
from taggit.models import ItemBase, TagBase, TaggedItemBase

from wagtail.admin import widgets
from wagtail.admin.forms import WagtailAdminPageForm
from wagtail.admin.forms.pages import CopyForm
from wagtail.admin.mail import send_mail
from wagtail.admin.panels import (
    FieldPanel,
    HelpPanel,
    InlinePanel,
    MultiFieldPanel,
    MultipleChooserPanel,
    ObjectList,
    PublishingPanel,
    TabbedInterface,
    TitleFieldPanel,
)
from wagtail.blocks import (
    CharBlock,
    FieldBlock,
    ListBlock,
    RawHTMLBlock,
    RichTextBlock,
    StreamBlock,
    StructBlock,
)
from wagtail.compat import HTTPMethod
from wagtail.contrib.forms.forms import FormBuilder, WagtailAdminFormPageForm
from wagtail.contrib.forms.models import (
    FORM_FIELD_CHOICES,
    AbstractEmailForm,
    AbstractFormField,
    AbstractFormSubmission,
)
from wagtail.contrib.forms.panels import FormSubmissionsPanel
from wagtail.contrib.forms.views import SubmissionsListView
from wagtail.contrib.settings.models import (
    BaseGenericSetting,
    BaseSiteSetting,
    register_setting,
)
from wagtail.contrib.sitemaps import Sitemap
from wagtail.contrib.table_block.blocks import TableBlock
from wagtail.documents import get_document_model
from wagtail.documents.blocks import DocumentChooserBlock
from wagtail.documents.models import AbstractDocument, Document
from wagtail.fields import RichTextField, StreamField
from wagtail.images import get_image_model
from wagtail.images.blocks import ImageBlock, ImageChooserBlock
from wagtail.images.models import AbstractImage, AbstractRendition, Image
from wagtail.models import (
    DraftStateMixin,
    LockableMixin,
    Orderable,
    Page,
    PageManager,
    PagePermissionTester,
    PageQuerySet,
    PreviewableMixin,
    RevisionMixin,
    Task,
    TaskState,
    TranslatableMixin,
    WorkflowMixin,
)
from wagtail.search import index
from wagtail.snippets.blocks import SnippetChooserBlock
from wagtail.snippets.models import register_snippet

from .forms import FormClassAdditionalFieldPageForm, ValidatedPageForm

EVENT_AUDIENCE_CHOICES = (
    ("public", _("Public")),
    ("private", _("Private")),
)


COMMON_PANELS = ("slug", "seo_title", "show_in_menus", "search_description")

CUSTOM_PREVIEW_SIZES = [
    {
        "name": "custom-mobile",
        "icon": "mobile-alt",
        "device_width": 412,
        "label": "Custom mobile preview",
    },
    {
        "name": "desktop",
        "icon": "desktop",
        "device_width": 1280,
        "label": "Original desktop",
    },
]


# Link fields


class LinkFields(models.Model):
    link_external = models.URLField("External link", blank=True)
    link_page = models.ForeignKey(
        "wagtailcore.Page",
        null=True,
        blank=True,
        related_name="+",
        on_delete=models.CASCADE,
    )
    link_document = models.ForeignKey(
        "wagtaildocs.Document",
        null=True,
        blank=True,
        related_name="+",
        on_delete=models.CASCADE,
    )

    @property
    def link(self):
        if self.link_page:
            return self.link_page.url
        elif self.link_document:
            return self.link_document.url
        else:
            return self.link_external

    panels = [
        FieldPanel("link_external"),
        FieldPanel("link_page"),
        FieldPanel("link_document"),
    ]

    class Meta:
        abstract = True


# Carousel items


class CarouselItem(LinkFields):
    image = models.ForeignKey(
        "wagtailimages.Image",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )
    embed_url = models.URLField("Embed URL", blank=True)
    caption = models.CharField(max_length=255, blank=True)

    panels = [
        "image",
        "embed_url",
        "caption",
        MultiFieldPanel(LinkFields.panels, "Link"),
    ]

    class Meta:
        abstract = True


# Related links


class RelatedLink(LinkFields):
    title = models.CharField(max_length=255, help_text="Link title")

    panels = [
        "title",
        MultiFieldPanel(LinkFields.panels, "Link"),
    ]

    class Meta:
        abstract = True


# Simple page
class SimplePage(Page):
    content = models.TextField()
    page_description = "A simple page description"

    content_panels = [
        TitleFieldPanel("title", classname="title"),
        FieldPanel("content"),
    ]

    def get_admin_display_title(self):
        return "%s (simple page)" % super().get_admin_display_title()


class MultiPreviewModesPage(Page):
    preview_templates = {
        "original": "tests/simple_page.html",
        "alt#1": "tests/simple_page_alt.html",
    }
    template = preview_templates["original"]

    @property
    def preview_modes(self):
        return [("original", "Original"), ("alt#1", "Alternate")]

    @property
    def default_preview_mode(self):
        return "alt#1"

    def get_preview_template(self, request, mode_name):
        if mode_name in self.preview_templates:
            return self.preview_templates[mode_name]
        return super().get_preview_template(request, mode_name)


class CustomPreviewSizesPage(Page):
    template = "tests/simple_page.html"

    @property
    def preview_sizes(self):
        return CUSTOM_PREVIEW_SIZES

    @property
    def default_preview_size(self):
        return "desktop"


class ExcludedCopyPageNote(Orderable):
    page = ParentalKey(
        "tests.PageWithExcludedCopyField",
        related_name="special_notes",
        on_delete=models.CASCADE,
    )
    note = models.CharField(max_length=255)

    panels = [FieldPanel("note")]


# Page with Excluded Fields when copied
class PageWithExcludedCopyField(Page):
    content = models.TextField()

    # Exclude these fields and the special_notes relation from being copied
    special_field = models.CharField(blank=True, max_length=255, default="Very Special")
    special_stream = StreamField(
        [("item", CharBlock())], default=[("item", "default item")]
    )
    exclude_fields_in_copy = ["special_field", "special_notes", "special_stream"]

    content_panels = [
        TitleFieldPanel("title", classname="title"),
        FieldPanel("special_field"),
        FieldPanel("content"),
        FieldPanel("special_stream"),
        InlinePanel("special_notes", label="special note"),
    ]


class RelatedGenericRelation(models.Model):
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveBigIntegerField()
    content_object = GenericForeignKey("content_type", "object_id")


class PageWithGenericRelation(Page):
    generic_relation = GenericRelation("tests.RelatedGenericRelation")


class PageWithOldStyleRouteMethod(Page):
    """
    Prior to Wagtail 0.4, the route() method on Page returned an HttpResponse
    rather than a Page instance. As subclasses of Page may override route,
    we need to continue accepting this convention (albeit as a deprecated API).
    """

    content = models.TextField()
    template = "tests/simple_page.html"

    def route(self, request, path_components):
        return self.serve(request)


# File page
class FilePage(Page):
    file_field = models.FileField()

    content_panels = [
        TitleFieldPanel("title", classname="title"),
        HelpPanel("remember to check for viruses"),
        FieldPanel("file_field"),
    ]


# Event page


class EventPageCarouselItem(TranslatableMixin, Orderable, CarouselItem):
    page = ParentalKey(
        "tests.EventPage", related_name="carousel_items", on_delete=models.CASCADE
    )

    class Meta(TranslatableMixin.Meta, Orderable.Meta):
        pass


class EventPageRelatedLink(TranslatableMixin, Orderable, RelatedLink):
    page = ParentalKey(
        "tests.EventPage", related_name="related_links", on_delete=models.CASCADE
    )

    class Meta(TranslatableMixin.Meta, Orderable.Meta):
        pass


class EventPageSpeakerAward(TranslatableMixin, Orderable, models.Model):
    speaker = ParentalKey(
        "tests.EventPageSpeaker", related_name="awards", on_delete=models.CASCADE
    )
    name = models.CharField("Award name", max_length=255)
    date_awarded = models.DateField(null=True, blank=True)

    panels = ["name", "date_awarded"]

    class Meta(TranslatableMixin.Meta, Orderable.Meta):
        pass


class EventPageSpeaker(TranslatableMixin, Orderable, LinkFields, ClusterableModel):
    page = ParentalKey(
        "tests.EventPage",
        related_name="speakers",
        related_query_name="speaker",
        on_delete=models.CASCADE,
    )
    first_name = models.CharField("Name", max_length=255, blank=True)
    last_name = models.CharField("Surname", max_length=255, blank=True)
    image = models.ForeignKey(
        "wagtailimages.Image",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )

    @property
    def name_display(self):
        return self.first_name + " " + self.last_name

    panels = [
        "first_name",
        "last_name",
        "image",
        MultiFieldPanel(LinkFields.panels, "Link"),
        InlinePanel("awards", label="award"),
    ]

    class Meta(TranslatableMixin.Meta, Orderable.Meta):
        pass


class EventCategory(TranslatableMixin, models.Model):
    name = models.CharField("Name", max_length=255)

    def __str__(self):
        return self.name


# Override the standard WagtailAdminPageForm to add validation on start/end dates
# that appears as a non-field error


class EventPageForm(WagtailAdminPageForm):
    def clean(self):
        cleaned_data = super().clean()

        # Make sure that the event starts before it ends
        start_date = cleaned_data.get("date_from")
        end_date = cleaned_data.get("date_to")
        if start_date and end_date and start_date > end_date:
            raise ValidationError("The end date must be after the start date")

        return cleaned_data


class EventPage(Page):
    date_from = models.DateField("Start date", null=True)
    date_to = models.DateField(
        "End date",
        null=True,
        blank=True,
        help_text="Not required if event is on a single day",
    )
    time_from = models.TimeField("Start time", null=True, blank=True)
    time_to = models.TimeField("End time", null=True, blank=True)
    audience = models.CharField(max_length=255, choices=EVENT_AUDIENCE_CHOICES)
    location = models.CharField(max_length=255)
    body = RichTextField(blank=True)
    cost = models.CharField(max_length=255)
    signup_link = models.URLField(blank=True)
    feed_image = models.ForeignKey(
        "wagtailimages.Image",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )
    categories = ParentalManyToManyField(EventCategory, blank=True)

    search_fields = Page.search_fields + [
        index.SearchField("get_audience_display"),
        index.SearchField("location"),
        index.SearchField("body"),
        index.FilterField("url_path"),
    ]

    password_required_template = "tests/event_page_password_required.html"
    base_form_class = EventPageForm

    content_panels = [
        TitleFieldPanel("title", classname="title"),
        "date_from",
        "date_to",
        "time_from",
        "time_to",
        "location",
        FieldPanel("audience", help_text="Who this event is for"),
        "cost",
        "signup_link",
        InlinePanel("carousel_items", label="carousel item"),
        "body",
        InlinePanel(
            "speakers",
            label="speaker",
            heading="Speaker lineup",
            help_text="Put the keynote speaker first",
        ),
        InlinePanel("related_links", label="related link"),
        "categories",
        # InlinePanel related model uses `pk` not `id`
        InlinePanel("head_counts", label="head count"),
    ]

    promote_panels = [
        MultiFieldPanel(
            COMMON_PANELS, "Common page configuration", help_text="For SEO nerds only"
        ),
        FieldPanel("feed_image"),
    ]

    class Meta:
        permissions = [
            ("custom_see_panel_setting", "Can see the panel."),
            ("other_custom_see_panel_setting", "Can see the panel."),
        ]


class HeadCountRelatedModelUsingPK(models.Model):
    """Related model that uses a custom primary key (pk) not id"""

    custom_id = models.AutoField(primary_key=True)
    event_page = ParentalKey(
        EventPage, on_delete=models.CASCADE, related_name="head_counts"
    )
    related_page = models.ForeignKey(
        Page,
        on_delete=models.CASCADE,
        related_name="head_count_relations",
        null=True,
    )
    head_count = models.IntegerField()
    panels = [FieldPanel("head_count")]


# Override the standard WagtailAdminPageForm to add field that is not in model
# so that we can test additional potential issues like comparing versions
class FormClassAdditionalFieldPage(Page):
    location = models.CharField(max_length=255)
    body = RichTextField(blank=True)

    content_panels = [
        TitleFieldPanel("title", classname="title"),
        FieldPanel("location"),
        FieldPanel("body"),
        FieldPanel("code"),  # not in model, see set base_form_class
    ]

    base_form_class = FormClassAdditionalFieldPageForm


# Just to be able to test multi table inheritance
class SingleEventPage(EventPage):
    excerpt = models.TextField(
        max_length=255,
        blank=True,
        null=True,
        help_text="Short text to describe what is this action about",
    )

    # Give this page model a custom URL routing scheme
    def get_url_parts(self, request=None):
        url_parts = super().get_url_parts(request=request)
        if url_parts is None:
            return None
        else:
            site_id, root_url, page_path = url_parts
            return (site_id, root_url, page_path + "pointless-suffix/")

    def route(self, request, path_components):
        if path_components == ["pointless-suffix"]:
            # treat this as equivalent to a request for this page
            return super().route(request, [])
        else:
            # fall back to default routing rules
            return super().route(request, path_components)

    def get_admin_display_title(self):
        return "%s (single event)" % super().get_admin_display_title()

    content_panels = [FieldPanel("excerpt")] + EventPage.content_panels


# "custom" sitemap object
class EventSitemap(Sitemap):
    pass


# Event index (has a separate AJAX template, and a custom template context)
class EventIndex(Page):
    intro = RichTextField(blank=True, max_length=50)
    ajax_template = "tests/includes/event_listing.html"

    # NOTE: Using a mix of enum and string values to test handling of both
    allowed_http_methods = [HTTPMethod.GET, "OPTIONS"]

    def get_events(self):
        return self.get_children().live().type(EventPage)

    def get_paginator(self):
        return Paginator(self.get_events(), 4)

    def get_context(self, request, page=1):
        # Pagination
        paginator = self.get_paginator()
        try:
            events = paginator.page(page)
        except PageNotAnInteger:
            events = paginator.page(1)
        except EmptyPage:
            events = paginator.page(paginator.num_pages)

        # Update context
        context = super().get_context(request)
        context["events"] = events
        return context

    def route(self, request, path_components):
        if self.live and len(path_components) == 1:
            try:
                return self.serve(request, page=int(path_components[0]))
            except (TypeError, ValueError):
                pass

        return super().route(request, path_components)

    def get_sitemap_urls(self, request=None):
        # Add past events url to sitemap
        return super().get_sitemap_urls(request=request) + [
            {
                "location": self.full_url + "past/",
                "lastmod": self.latest_revision_created_at,
            }
        ]

    def get_cached_paths(self):
        return super().get_cached_paths() + ["/past/"]

    content_panels = [
        TitleFieldPanel("title", classname="title"),
        FieldPanel("intro"),
    ]


class FormField(AbstractFormField):
    page = ParentalKey("FormPage", related_name="form_fields", on_delete=models.CASCADE)


class FormPage(AbstractEmailForm):
    def get_context(self, request):
        context = super().get_context(request)
        context["greeting"] = "hello world"
        return context

    # This is redundant (SubmissionsListView is the default view class), but importing
    # SubmissionsListView in this models.py helps us to confirm that this recipe
    # https://docs.wagtail.org/en/stable/reference/contrib/forms/customization.html#customise-form-submissions-listing-in-wagtail-admin
    # works without triggering circular dependency issues -
    # see https://github.com/wagtail/wagtail/issues/6265
    submissions_list_view_class = SubmissionsListView

    content_panels = [
        TitleFieldPanel("title", classname="title"),
        InlinePanel("form_fields", label="form field"),
        MultiFieldPanel(
            [
                FieldPanel("to_address"),
                FieldPanel("from_address"),
                FieldPanel("subject"),
            ],
            "Email",
        ),
        FormSubmissionsPanel(),
    ]


# CopyForm allowing auto-increment of slugs


class CustomCopyForm(CopyForm):
    def __init__(self, *args, **kwargs):
        # call super
        super().__init__(*args, **kwargs)
        # set initial_slug as incremented slug
        suffix = 2
        parent_page = self.page.get_parent()
        if self.page.slug:
            try:
                suffix = int(self.page.slug[-1]) + 1
                base_slug = self.page.slug[:-2]

            except ValueError:
                base_slug = self.page.slug

        candidate_slug = base_slug + f"-{suffix}"
        while not Page._slug_is_available(candidate_slug, parent_page):
            suffix += 1
            candidate_slug = f"{base_slug}-{suffix}"
            candidate_slug
        allow_unicode = getattr(settings, "WAGTAIL_ALLOW_UNICODE_SLUGS", True)
        self.fields["new_slug"] = forms.SlugField(
            initial=candidate_slug,
            label=_("New slug"),
            allow_unicode=allow_unicode,
            widget=widgets.slug.SlugInput,
        )


# FormPage with a non-HTML extension


class JadeFormField(AbstractFormField):
    page = ParentalKey(
        "JadeFormPage", related_name="form_fields", on_delete=models.CASCADE
    )


class JadeFormPage(AbstractEmailForm):
    template = "tests/form_page.jade"

    content_panels = [
        TitleFieldPanel("title", classname="title"),
        InlinePanel("form_fields", label="form field"),
        MultiFieldPanel(
            [
                FieldPanel("to_address"),
                FieldPanel("from_address"),
                FieldPanel("subject"),
            ],
            "Email",
        ),
    ]


# Form page that redirects to a different page


class RedirectFormField(AbstractFormField):
    page = ParentalKey(
        "FormPageWithRedirect", related_name="form_fields", on_delete=models.CASCADE
    )


class FormPageWithRedirect(AbstractEmailForm):
    thank_you_redirect_page = models.ForeignKey(
        "wagtailcore.Page",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )

    def get_context(self, request):
        context = super().get_context(request)
        context["greeting"] = "hello world"
        return context

    def render_landing_page(self, request, form_submission=None, *args, **kwargs):
        """
        Renders the landing page OR if a receipt_page_redirect is chosen redirects to this page.
        """
        if self.thank_you_redirect_page:
            return redirect(self.thank_you_redirect_page.url, permanent=False)

        return super().render_landing_page(request, form_submission, *args, **kwargs)

    content_panels = [
        TitleFieldPanel("title", classname="title"),
        FieldPanel("thank_you_redirect_page"),
        InlinePanel("form_fields", label="form field"),
        MultiFieldPanel(
            [
                FieldPanel("to_address"),
                FieldPanel("from_address"),
                FieldPanel("subject"),
            ],
            "Email",
        ),
    ]


# FormPage with a custom FormSubmission


class FormPageWithCustomSubmissionForm(WagtailAdminFormPageForm):
    """
    Used to validate that admin forms can validate the page's submissions via
    extending the form class.
    """

    def clean(self):
        cleaned_data = super().clean()
        from_address = cleaned_data.get("from_address")
        if from_address and "example.com" in from_address:
            raise ValidationError("Email cannot be from example.com")

        return cleaned_data


class FormPageWithCustomSubmission(AbstractEmailForm):
    """
    A ``FormPage`` with a custom FormSubmission and other extensive customizations:

    * A custom submission model
    * A custom related_name (see `FormFieldWithCustomSubmission.page`)
    * Saves reference to a user
    * Doesn't render html form, if submission for current user is present
    * A custom clean method that does not allow the ``from_address`` to be set to anything including example.com
    """

    intro = RichTextField(blank=True)
    thank_you_text = RichTextField(blank=True)

    base_form_class = FormPageWithCustomSubmissionForm

    def get_context(self, request, *args, **kwargs):
        context = super().get_context(request)
        context["greeting"] = "hello world"
        return context

    def get_form_fields(self):
        return self.custom_form_fields.all()

    def get_data_fields(self):
        data_fields = [
            ("useremail", "User email"),
        ]
        data_fields += super().get_data_fields()

        return data_fields

    def get_submission_class(self):
        return CustomFormPageSubmission

    def process_form_submission(self, form):
        form_submission = self.get_submission_class().objects.create(
            form_data=form.cleaned_data,
            page=self,
            user=form.user,
        )

        if self.to_address:
            addresses = [x.strip() for x in self.to_address.split(",")]
            content = "\n".join(
                [
                    x[1].label + ": " + str(form.data.get(x[0]))
                    for x in form.fields.items()
                ]
            )
            send_mail(
                self.subject,
                content,
                addresses,
                self.from_address,
            )

        # process_form_submission should now return the created form_submission
        return form_submission

    def serve(self, request, *args, **kwargs):
        if (
            self.get_submission_class()
            .objects.filter(page=self, user__pk=request.user.pk)
            .exists()
        ):
            return TemplateResponse(request, self.template, self.get_context(request))

        return super().serve(request, *args, **kwargs)

    content_panels = [
        TitleFieldPanel("title", classname="title"),
        FieldPanel("intro"),
        InlinePanel("custom_form_fields", label="form field"),
        FieldPanel("thank_you_text"),
        MultiFieldPanel(
            [
                FieldPanel("to_address"),
                FieldPanel("from_address"),
                FieldPanel("subject"),
            ],
            "Email",
        ),
    ]


class FormFieldWithCustomSubmission(AbstractFormField):
    page = ParentalKey(
        FormPageWithCustomSubmission,
        on_delete=models.CASCADE,
        related_name="custom_form_fields",
    )


class CustomFormPageSubmission(AbstractFormSubmission):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)

    def get_data(self):
        form_data = super().get_data()
        form_data.update(
            {
                "useremail": self.user.email,
            }
        )

        return form_data


# Custom form page with custom submission listing view and form submission


class FormFieldForCustomListViewPage(AbstractFormField):
    page = ParentalKey(
        "FormPageWithCustomSubmissionListView",
        related_name="form_fields",
        on_delete=models.CASCADE,
    )


class FormPageWithCustomSubmissionListView(AbstractEmailForm):
    """Form Page with customised submissions listing view"""

    intro = RichTextField(blank=True)
    thank_you_text = RichTextField(blank=True)

    def get_submissions_list_view_class(self):
        from .views import CustomSubmissionsListView

        return CustomSubmissionsListView

    def get_submission_class(self):
        return CustomFormPageSubmission

    def get_data_fields(self):
        data_fields = [
            ("useremail", "User email"),
        ]
        data_fields += super().get_data_fields()

        return data_fields

    content_panels = [
        TitleFieldPanel("title", classname="title"),
        FieldPanel("intro"),
        InlinePanel("form_fields", label="form field"),
        FieldPanel("thank_you_text"),
        MultiFieldPanel(
            [
                FieldPanel("to_address"),
                FieldPanel("from_address"),
                FieldPanel("subject"),
            ],
            "Email",
        ),
    ]


# FormPage with custom FormBuilder

EXTENDED_CHOICES = FORM_FIELD_CHOICES + (("ipaddress", "IP Address"),)


class ExtendedFormField(AbstractFormField):
    """
    Override the field_type field with extended choices
    and a custom clean_name override.
    """

    page = ParentalKey(
        "FormPageWithCustomFormBuilder",
        related_name="form_fields",
        on_delete=models.CASCADE,
    )
    field_type = models.CharField(
        verbose_name="field type", max_length=16, choices=EXTENDED_CHOICES
    )

    def get_field_clean_name(self):
        clean_name = super().get_field_clean_name()

        # scoping to field type to easily test behaviour in isolation
        if self.field_type == "number":
            return f"number_field--{clean_name}"

        # scoping to field label to easily test duplicate behaviour in isolation
        if "duplicate" in self.label:
            return "test duplicate"

        return clean_name


class CustomFormBuilder(FormBuilder):
    """
    A custom FormBuilder that has an 'ipaddress' field with
    customised create_singleline_field with shorter max_length
    """

    def create_singleline_field(self, field, options):
        options["max_length"] = 120  # usual default is 255
        return forms.CharField(**options)

    def create_ipaddress_field(self, field, options):
        return forms.GenericIPAddressField(**options)


class FormBuilderWithCustomWidget(FormBuilder):
    """
    A form builder that customizes all default field type and
    passes a `widget` parameter in the options to the parent class.
    """

    def create_singleline_field(self, field, options):
        options["widget"] = forms.TextInput(attrs={"class": "custom"})
        return super().create_singleline_field(field, options)

    def create_multiline_field(self, field, options):
        options["widget"] = forms.TextInput(attrs={"class": "custom"})
        return super().create_multiline_field(field, options)

    def create_email_field(self, field, options):
        options["widget"] = forms.TextInput(attrs={"class": "custom"})
        return super().create_email_field(field, options)

    def create_number_field(self, field, options):
        options["widget"] = forms.TextInput(attrs={"class": "custom"})
        return super().create_number_field(field, options)

    def create_url_field(self, field, options):
        options["widget"] = forms.TextInput(attrs={"class": "custom"})
        return super().create_url_field(field, options)

    def create_checkbox_field(self, field, options):
        options["widget"] = forms.TextInput(attrs={"class": "custom"})
        return super().create_checkbox_field(field, options)

    def create_checkboxes_field(self, field, options):
        options["widget"] = forms.TextInput(attrs={"class": "custom"})
        return super().create_checkboxes_field(field, options)

    def create_dropdown_field(self, field, options):
        options["widget"] = forms.TextInput(attrs={"class": "custom"})
        return super().create_dropdown_field(field, options)

    def create_multiselect_field(self, field, options):
        options["widget"] = forms.TextInput(attrs={"class": "custom"})
        return super().create_multiselect_field(field, options)

    def create_radio_field(self, field, options):
        options["widget"] = forms.TextInput(attrs={"class": "custom"})
        return super().create_radio_field(field, options)

    def create_date_field(self, field, options):
        options["widget"] = forms.TextInput(attrs={"class": "custom"})
        return super().create_date_field(field, options)

    def create_datetime_field(self, field, options):
        options["widget"] = forms.TextInput(attrs={"class": "custom"})
        return super().create_datetime_field(field, options)

    def create_hidden_field(self, field, options):
        options["widget"] = forms.TextInput(attrs={"class": "custom"})
        return super().create_hidden_field(field, options)


class FormPageWithCustomFormBuilder(AbstractEmailForm):
    """
    A Form page that has a custom form builder and uses a custom
    form field model with additional field_type choices.
    """

    form_builder = CustomFormBuilder

    content_panels = [
        TitleFieldPanel("title", classname="title"),
        InlinePanel("form_fields", label="form field"),
        MultiFieldPanel(
            [
                FieldPanel("to_address"),
                FieldPanel("from_address"),
                FieldPanel("subject"),
            ],
            "Email",
        ),
    ]


# Snippets
class AdvertPlacement(models.Model):
    page = ParentalKey(
        "wagtailcore.Page", related_name="advert_placements", on_delete=models.CASCADE
    )
    advert = models.ForeignKey(
        "tests.Advert", related_name="+", on_delete=models.CASCADE
    )
    colour = models.CharField(max_length=255)


class AdvertTag(TaggedItemBase):
    content_object = ParentalKey(
        "Advert", related_name="tagged_items", on_delete=models.CASCADE
    )


class Advert(ClusterableModel):
    url = models.URLField(null=True, blank=True)
    text = models.CharField(max_length=255)

    tags = TaggableManager(through=AdvertTag, blank=True)

    panels = [
        FieldPanel("url"),
        FieldPanel("text"),
        FieldPanel("tags"),
    ]

    def __str__(self):
        return self.text


register_snippet(Advert)


class AdvertWithCustomPrimaryKey(ClusterableModel):
    advert_id = models.CharField(max_length=255, primary_key=True)
    url = models.URLField(null=True, blank=True)
    text = models.CharField(max_length=255)

    panels = [
        FieldPanel("url"),
        FieldPanel("text"),
    ]

    def __str__(self):
        return self.text


register_snippet(AdvertWithCustomPrimaryKey)


class AdvertWithCustomUUIDPrimaryKey(index.Indexed, ClusterableModel):
    advert_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    url = models.URLField(null=True, blank=True)
    text = models.CharField(max_length=255)
    page = models.ForeignKey(Page, null=True, blank=True, on_delete=models.SET_NULL)

    panels = [
        FieldPanel("url"),
        FieldPanel("text"),
        FieldPanel("page"),
    ]

    search_fields = [
        index.SearchField("text"),
    ]

    def __str__(self):
        return self.text


register_snippet(AdvertWithCustomUUIDPrimaryKey)


class AdvertWithTabbedInterface(models.Model):
    url = models.URLField(null=True, blank=True)
    text = models.CharField(max_length=255)
    something_else = models.CharField(max_length=255)

    advert_panels = [
        FieldPanel("url"),
        FieldPanel("text"),
    ]

    other_panels = [
        FieldPanel("something_else"),
    ]

    edit_handler = TabbedInterface(
        [
            ObjectList(advert_panels, heading="Advert"),
            ObjectList(
                other_panels, heading="Other", help_text="Other panels help text"
            ),
        ],
        help_text="Top-level help text",
    )

    def __str__(self):
        return self.text

    class Meta:
        ordering = ("text",)


register_snippet(AdvertWithTabbedInterface)


class CustomManager(models.Manager):
    pass


class ModelWithCustomManager(models.Model):
    instances = CustomManager()


register_snippet(ModelWithCustomManager)


# Models with RevisionMixin
class RevisableModel(RevisionMixin, models.Model):
    text = models.TextField()


class RevisableChildModel(RevisableModel):
    secret_text = models.TextField(blank=True, default="")

    # The edit_handler is defined on the viewset


class RevisableGrandChildModel(RevisableChildModel):
    pass


# Models with DraftStateMixin
class DraftStateModel(DraftStateMixin, LockableMixin, RevisionMixin, models.Model):
    text = models.TextField()

    # The panels are defined on the viewset

    def __str__(self):
        return self.text


class DraftStateCustomPrimaryKeyModel(DraftStateMixin, RevisionMixin, models.Model):
    custom_id = models.CharField(max_length=255, primary_key=True)
    text = models.TextField()

    panels = [
        FieldPanel("text"),
        FieldPanel("first_published_at"),
        PublishingPanel(),
    ]

    def __str__(self):
        return self.text


register_snippet(DraftStateCustomPrimaryKeyModel)


# Models with PreviewableMixin
class PreviewableModel(PreviewableMixin, ClusterableModel):
    text = models.TextField()
    categories = ParentalManyToManyField(EventCategory, blank=True)

    def __str__(self):
        return self.text

    def get_preview_template(self, request, mode_name):
        return "tests/previewable_model.html"


register_snippet(PreviewableModel)


class CustomPreviewSizesModel(PreviewableMixin, models.Model):
    text = models.TextField()

    @property
    def preview_sizes(self):
        return CUSTOM_PREVIEW_SIZES

    @property
    def default_preview_size(self):
        return "desktop"


register_snippet(CustomPreviewSizesModel)


class MultiPreviewModesModel(PreviewableMixin, RevisionMixin, models.Model):
    text = models.TextField()

    def __str__(self):
        return self.text

    @property
    def preview_modes(self):
        return [("", "Normal"), ("alt#1", "Alternate")]

    @property
    def default_preview_mode(self):
        return "alt#1"

    def get_preview_template(self, request, mode_name):
        templates = {
            "": "tests/previewable_model.html",
            "alt#1": "tests/previewable_model_alt.html",
        }
        return templates.get(mode_name, templates[""])


register_snippet(MultiPreviewModesModel)


class NonPreviewableModel(PreviewableMixin, RevisionMixin, models.Model):
    text = models.TextField()

    def __str__(self):
        return self.text

    preview_modes = []


register_snippet(NonPreviewableModel)


# Models with LockableMixin


class LockableModel(LockableMixin, models.Model):
    text = models.TextField()

    def __str__(self):
        return self.text


register_snippet(LockableModel)


# Models with WorkflowMixin
# Note: do not use Workflow in the model name to avoid incorrect counts in tests
# that look for the word "workflow"


class ModeratedModel(WorkflowMixin, DraftStateMixin, RevisionMixin, models.Model):
    text = models.TextField()

    def __str__(self):
        return self.text


# Snippet with all mixins enabled


class FullFeaturedSnippet(
    PreviewableMixin,
    WorkflowMixin,
    DraftStateMixin,
    LockableMixin,
    RevisionMixin,
    TranslatableMixin,
    index.Indexed,
    models.Model,
):
    class CountryCode(models.TextChoices):
        INDONESIA = "ID"
        PHILIPPINES = "PH"
        UNITED_KINGDOM = "UK"

    text = models.TextField()
    country_code = models.CharField(
        max_length=2,
        choices=CountryCode.choices,
        default=CountryCode.UNITED_KINGDOM,
        blank=True,
    )
    some_date = models.DateField(auto_now=True)
    some_number = models.IntegerField(default=0, blank=True)

    some_attribute = "some value"

    workflow_states = GenericRelation(
        "wagtailcore.WorkflowState",
        content_type_field="base_content_type",
        object_id_field="object_id",
        related_query_name="full_featured_snippet",
        for_concrete_model=False,
    )

    revisions = GenericRelation(
        "wagtailcore.Revision",
        content_type_field="base_content_type",
        object_id_field="object_id",
        related_query_name="full_featured_snippet",
        for_concrete_model=False,
    )

    panels = ["text", "country_code", "some_number"]

    search_fields = [
        index.SearchField("text"),
        index.AutocompleteField("text"),
        index.FilterField("text"),
        index.FilterField("country_code"),
    ]

    def __str__(self):
        return self.text

    def modulo_two(self):
        return self.pk % 2

    def tristate(self):
        return (None, True, False)[self.pk % 3]

    def get_preview_template(self, request, mode_name):
        return "tests/previewable_model.html"

    def get_foo_country_code(self):
        return f"Foo {self.country_code}"

    get_foo_country_code.admin_order_field = "country_code"
    get_foo_country_code.short_description = "custom FOO column"

    class Meta(TranslatableMixin.Meta):
        verbose_name = "full-featured snippet"
        verbose_name_plural = "full-featured snippets"


def get_default_advert():
    return Advert.objects.first()


class VariousOnDeleteModel(models.Model):
    text = models.TextField()
    on_delete_cascade = models.ForeignKey(
        Advert, on_delete=models.CASCADE, null=True, blank=True, related_name="+"
    )
    on_delete_protect = models.ForeignKey(
        Advert, on_delete=models.PROTECT, null=True, blank=True, related_name="+"
    )
    on_delete_restrict = models.ForeignKey(
        Advert, on_delete=models.RESTRICT, null=True, blank=True, related_name="+"
    )
    on_delete_set_null = models.ForeignKey(
        Advert, on_delete=models.SET_NULL, null=True, blank=True, related_name="+"
    )
    on_delete_set_default = models.ForeignKey(
        Advert,
        on_delete=models.SET_DEFAULT,
        null=True,
        blank=True,
        default=get_default_advert,
        related_name="+",
    )
    on_delete_set = models.ForeignKey(
        Advert,
        on_delete=models.SET(get_default_advert),
        null=True,
        blank=True,
        related_name="+",
    )
    on_delete_do_nothing = models.ForeignKey(
        Advert, on_delete=models.DO_NOTHING, null=True, blank=True, related_name="+"
    )

    protected_image = models.ForeignKey(
        "wagtailimages.Image",
        on_delete=models.PROTECT,
        null=True,
        blank=True,
        related_name="+",
    )
    protected_document = models.ForeignKey(
        "wagtaildocs.Document",
        on_delete=models.PROTECT,
        null=True,
        blank=True,
        related_name="+",
    )
    protected_page = models.ForeignKey(
        "wagtailcore.Page",
        on_delete=models.PROTECT,
        null=True,
        blank=True,
        related_name="+",
    )
    protected_user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        null=True,
        blank=True,
        related_name="+",
    )

    cascading_toy = models.ForeignKey(
        "tests.FeatureCompleteToy",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        related_name="+",
    )

    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, null=True, blank=True
    )
    object_id = models.UUIDField(null=True, blank=True)
    content_object = GenericForeignKey("content_type", "object_id")

    stream_field = StreamField(
        [
            (
                "advertisement_content",
                StreamBlock(
                    [
                        (
                            "captioned_advert",
                            StructBlock(
                                [
                                    ("advert", SnippetChooserBlock(Advert)),
                                    ("caption", CharBlock()),
                                ],
                            ),
                        ),
                        ("rich_text", RichTextBlock()),
                    ]
                ),
            ),
            ("image", ImageChooserBlock()),
            ("document", DocumentChooserBlock()),
        ],
    )
    rich_text = RichTextField(blank=True)


class StandardIndex(Page):
    """Index for the site"""

    parent_page_types = [Page]

    # A custom panel setup where all Promote fields are placed in the Content tab instead;
    # we use this to test that the 'promote' tab is left out of the output when empty
    content_panels = [
        TitleFieldPanel("title", classname="title"),
        FieldPanel("seo_title"),
        FieldPanel("slug"),
        InlinePanel("advert_placements", heading="Adverts", label="advert"),
    ]

    promote_panels = []


class PromotionalPage(Page):
    content_panels = Page.content_panels + [
        InlinePanel("advert_placements", heading="Adverts", label="advert", min_num=1),
    ]


class StandardChild(Page):
    pass


# Test overriding edit_handler with a custom one
StandardChild.edit_handler = TabbedInterface(
    [
        ObjectList(StandardChild.content_panels, heading="Content"),
        ObjectList(StandardChild.promote_panels, heading="Promote"),
        ObjectList(StandardChild.settings_panels, heading="Settings"),
        ObjectList(
            [
                HelpPanel("Watch out for asteroids"),
            ],
            heading="Dinosaurs",
        ),
    ],
    base_form_class=WagtailAdminPageForm,
)


class BusinessIndex(Page):
    """Can be placed anywhere, can only have Business children"""

    subpage_types = ["tests.BusinessChild", "tests.BusinessSubIndex"]


class BusinessSubIndex(Page):
    """Can be placed under BusinessIndex, and have BusinessChild children"""

    # BusinessNowherePage is 'incorrectly' added here as a possible child.
    # The rules on BusinessNowherePage prevent it from being a child here though.
    subpage_types = ["tests.BusinessChild", "tests.BusinessNowherePage"]
    parent_page_types = ["tests.BusinessIndex"]


class BusinessChild(Page):
    """Can only be placed under Business indexes, no children allowed"""

    subpage_types = []
    parent_page_types = ["tests.BusinessIndex", BusinessSubIndex]
    page_description = _("A lazy business child page description")


class BusinessNowherePage(Page):
    """Not allowed to be placed anywhere"""

    parent_page_types = []


class CustomCopyFormPage(Page):
    copy_form_class = CustomCopyForm


class TaggedPageTag(TaggedItemBase):
    content_object = ParentalKey(
        "tests.TaggedPage", related_name="tagged_items", on_delete=models.CASCADE
    )


class TaggedPage(Page):
    tags = ClusterTaggableManager(through=TaggedPageTag, blank=True)

    content_panels = [
        TitleFieldPanel("title", classname="title"),
        FieldPanel("tags"),
    ]

    # Page.search_fields intentionally omitted to test warning
    search_fields = [
        index.SearchField("tags"),
    ]


class TaggedChildPage(TaggedPage):
    pass


class TaggedGrandchildPage(TaggedChildPage):
    pass


class SingletonPage(Page):
    @classmethod
    def can_create_at(cls, parent):
        # You can only create one of these!
        return super().can_create_at(parent) and not cls.objects.exists()


class SingletonPageViaMaxCount(Page):
    max_count = 1


class PageChooserModel(models.Model):
    page = models.ForeignKey(
        "wagtailcore.Page", help_text="help text", on_delete=models.CASCADE
    )


class EventPageChooserModel(models.Model):
    page = models.ForeignKey(
        "tests.EventPage", help_text="more help text", on_delete=models.CASCADE
    )


class SnippetChooserModel(models.Model):
    advert = models.ForeignKey(Advert, help_text="help text", on_delete=models.CASCADE)
    full_featured = models.ForeignKey(
        FullFeaturedSnippet,
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        verbose_name="Chosen snippet",
    )

    panels = [
        FieldPanel("advert"),
        FieldPanel("full_featured"),
    ]


class SnippetChooserModelWithCustomPrimaryKey(models.Model):
    advertwithcustomprimarykey = models.ForeignKey(
        AdvertWithCustomPrimaryKey, help_text="help text", on_delete=models.CASCADE
    )

    panels = [
        FieldPanel("advertwithcustomprimarykey"),
    ]


class CustomImage(AbstractImage):
    caption = models.CharField(max_length=255, blank=True)
    fancy_caption = RichTextField(blank=True)
    not_editable_field = models.CharField(max_length=255, blank=True)

    admin_form_fields = Image.admin_form_fields + (
        "caption",
        "fancy_caption",
    )

    class Meta:
        unique_together = [("title", "collection")]


class CustomRendition(AbstractRendition):
    image = models.ForeignKey(
        CustomImage, related_name="renditions", on_delete=models.CASCADE
    )

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=("image", "filter_spec", "focal_point_key"),
                name="unique_rendition",
            )
        ]


# Custom image model with a required field
class CustomImageWithAuthor(AbstractImage):
    author = models.CharField(max_length=255)

    admin_form_fields = Image.admin_form_fields + ("author",)


class CustomRenditionWithAuthor(AbstractRendition):
    image = models.ForeignKey(
        CustomImageWithAuthor, related_name="renditions", on_delete=models.CASCADE
    )

    class Meta:
        unique_together = (("image", "filter_spec", "focal_point_key"),)


class CustomDocument(AbstractDocument):
    description = models.TextField(blank=True)
    fancy_description = RichTextField(blank=True)
    admin_form_fields = Document.admin_form_fields + (
        "description",
        "fancy_description",
    )

    class Meta:
        unique_together = [("title", "collection")]


# Custom document model with a required field
class CustomDocumentWithAuthor(AbstractDocument):
    author = models.CharField(max_length=255)

    admin_form_fields = Document.admin_form_fields + ("author",)


class JSONStreamModel(models.Model):
    body = StreamField(
        [
            ("text", CharBlock()),
            ("rich_text", RichTextBlock()),
            ("image", ImageChooserBlock()),
        ],
    )

    class Meta:
        verbose_name = "JSON stream model"


class JSONMinMaxCountStreamModel(models.Model):
    body = StreamField(
        [
            ("text", CharBlock()),
            ("rich_text", RichTextBlock()),
            ("image", ImageChooserBlock()),
        ],
        min_num=2,
        max_num=5,
    )


class JSONBlockCountsStreamModel(models.Model):
    body = StreamField(
        [
            ("text", CharBlock()),
            ("rich_text", RichTextBlock()),
            ("image", ImageChooserBlock()),
        ],
        block_counts={
            "text": {"min_num": 1},
            "rich_text": {"max_num": 1},
            "image": {"min_num": 1, "max_num": 1},
        },
    )


class ExtendedImageChooserBlock(ImageChooserBlock):
    """
    Example of Block with custom get_api_representation method.
    If the request has an 'extended' query param, it returns a dict of id and title,
    otherwise, it returns the default value.
    """

    def get_api_representation(self, value, context=None):
        image_id = super().get_api_representation(value, context=context)
        if "request" in context and context["request"].query_params.get(
            "extended", False
        ):
            return {"id": image_id, "title": value.title}
        return image_id


class StreamPage(Page):
    body = StreamField(
        [
            ("text", CharBlock()),
            ("rich_text", RichTextBlock()),
            ("image", ExtendedImageChooserBlock()),
            (
                "product",
                StructBlock(
                    [
                        ("name", CharBlock()),
                        ("price", CharBlock()),
                    ]
                ),
            ),
            ("raw_html", RawHTMLBlock()),
            (
                "books",
                StreamBlock(
                    [
                        ("title", CharBlock()),
                        ("author", CharBlock()),
                    ]
                ),
            ),
            (
                "title_list",
                ListBlock(CharBlock()),
            ),
            ("image_with_alt", ImageBlock()),
        ],
    )

    api_fields = ("body",)

    content_panels = [
        TitleFieldPanel("title"),
        FieldPanel("body"),
    ]

    preview_modes = []


class DefaultStreamPage(Page):
    body = StreamField(
        [
            ("text", CharBlock()),
            ("rich_text", RichTextBlock()),
            ("image", ImageChooserBlock()),
        ],
        default="",
    )

    content_panels = [
        TitleFieldPanel("title"),
        FieldPanel("body"),
    ]


class ComplexDefaultStreamPage(Page):
    body = StreamField(
        [
            ("text", CharBlock()),
            ("rich_text", RichTextBlock()),
            (
                "books",
                StreamBlock(
                    [
                        ("title", CharBlock()),
                        ("author", CharBlock()),
                    ]
                ),
            ),
        ],
        default=[
            ("rich_text", "<p>My <i>lovely</i> books</p>"),
            (
                "books",
                [("title", "The Great Gatsby"), ("author", "F. Scott Fitzgerald")],
            ),
        ],
    )

    content_panels = [
        TitleFieldPanel("title"),
        FieldPanel("body"),
    ]


class MTIBasePage(Page):
    is_creatable = False

    class Meta:
        verbose_name = "MTI base page"


class MTIChildPage(MTIBasePage):
    # Should be creatable by default, no need to set anything
    pass


class NoCreatableSubpageTypesPage(Page):
    subpage_types = [MTIBasePage]


class NoSubpageTypesPage(Page):
    subpage_types = []


class AbstractPage(Page):
    class Meta:
        abstract = True


@register_setting
class TestSiteSetting(BaseSiteSetting):
    title = models.CharField(max_length=100)
    email = models.EmailField(max_length=50)


@register_setting
class TestGenericSetting(BaseGenericSetting):
    title = models.CharField(max_length=100)
    email = models.EmailField(max_length=50)


@register_setting
class TestPermissionedGenericSetting(BaseGenericSetting):
    title = models.CharField(max_length=100)
    sensitive_email = models.EmailField(max_length=50)

    panels = [
        FieldPanel("title"),
        FieldPanel(
            "sensitive_email",
            permission="tests.can_edit_sensitive_email_generic_setting",
        ),
    ]

    class Meta:
        permissions = [
            (
                "can_edit_sensitive_email_generic_setting",
                "Can edit sensitive email generic setting.",
            ),
        ]


@register_setting
class TestPermissionedSiteSetting(BaseSiteSetting):
    title = models.CharField(max_length=100)
    sensitive_email = models.EmailField(max_length=50)

    panels = [
        FieldPanel("title"),
        FieldPanel(
            "sensitive_email", permission="tests.can_edit_sensitive_email_site_setting"
        ),
    ]

    class Meta:
        permissions = [
            (
                "can_edit_sensitive_email_site_setting",
                "Can edit sensitive email site setting.",
            ),
        ]


@register_setting
class ImportantPagesSiteSetting(BaseSiteSetting):
    sign_up_page = models.ForeignKey(
        "wagtailcore.Page", related_name="+", null=True, on_delete=models.SET_NULL
    )
    general_terms_page = models.ForeignKey(
        "wagtailcore.Page", related_name="+", null=True, on_delete=models.SET_NULL
    )
    privacy_policy_page = models.ForeignKey(
        "wagtailcore.Page", related_name="+", null=True, on_delete=models.SET_NULL
    )


@register_setting(name="important-pages-generic-setting")
class ImportantPagesGenericSetting(BaseGenericSetting):
    sign_up_page = models.ForeignKey(
        "wagtailcore.Page", related_name="+", null=True, on_delete=models.SET_NULL
    )
    general_terms_page = models.ForeignKey(
        "wagtailcore.Page", related_name="+", null=True, on_delete=models.SET_NULL
    )
    privacy_policy_page = models.ForeignKey(
        "wagtailcore.Page", related_name="+", null=True, on_delete=models.SET_NULL
    )

    class Meta:
        verbose_name = _("important pages settings")
        verbose_name_plural = _("important pages settings")


@register_setting(icon="tag")
class IconSiteSetting(BaseSiteSetting):
    pass


@register_setting(icon="tag")
class IconGenericSetting(BaseGenericSetting):
    pass


@register_setting
class FileSiteSetting(BaseSiteSetting):
    file = models.FileField()


@register_setting
class FileGenericSetting(BaseGenericSetting):
    file = models.FileField()


@register_setting
class PreviewableSiteSetting(PreviewableMixin, BaseSiteSetting):
    text = models.TextField()

    def get_preview_template(self, request, mode_name):
        return "tests/previewable_setting.html"


@register_setting
class PreviewableGenericSetting(PreviewableMixin, BaseGenericSetting):
    text = models.TextField()

    def get_preview_template(self, request, mode_name):
        return "tests/previewable_setting.html"


class BlogCategory(models.Model):
    name = models.CharField(unique=True, max_length=80)


class BlogCategoryBlogPage(models.Model):
    category = models.ForeignKey(
        BlogCategory, related_name="+", on_delete=models.CASCADE
    )
    page = ParentalKey(
        "ManyToManyBlogPage", related_name="categories", on_delete=models.CASCADE
    )
    panels = [
        FieldPanel("category"),
    ]


class ManyToManyBlogPage(Page):
    """
    A page type with two different kinds of M2M relation.
    We don't formally support these, but we don't want them to cause
    hard breakages either.
    """

    body = RichTextField(blank=True)
    adverts = models.ManyToManyField(Advert, blank=True)
    blog_categories = models.ManyToManyField(
        BlogCategory, through=BlogCategoryBlogPage, blank=True
    )

    # make first_published_at editable on this page model
    settings_panels = Page.settings_panels + [
        FieldPanel("first_published_at"),
    ]


class OneToOnePage(Page):
    """
    A Page containing a O2O relation.
    """

    body = RichTextBlock(blank=True)
    page_ptr = models.OneToOneField(
        Page, parent_link=True, related_name="+", on_delete=models.CASCADE
    )


class GenericSnippetPage(Page):
    """
    A page containing a reference to an arbitrary snippet (or any model for that matter)
    linked by a GenericForeignKey
    """

    snippet_content_type = models.ForeignKey(
        ContentType, on_delete=models.SET_NULL, null=True, blank=True
    )
    snippet_object_id = models.PositiveIntegerField(null=True, blank=True)
    snippet_content_object = GenericForeignKey(
        "snippet_content_type", "snippet_object_id"
    )


class CustomImageFilePath(AbstractImage):
    def get_upload_to(self, filename):
        """Create a path that's file-system friendly.

        By hashing the file's contents we guarantee an equal distribution
        of files within our root directories. This also gives us a
        better chance of uploading images with the same filename, but
        different contents - this isn't guaranteed as we're only using
        the first three characters of the checksum.
        """
        original_filepath = super().get_upload_to(filename)
        folder_name, filename = original_filepath.split(os.path.sep)

        # Ensure that we consume the entire file, we can't guarantee that
        # the stream has not be partially (or entirely) consumed by
        # another process
        original_position = self.file.tell()
        self.file.seek(0)
        hash256 = hashlib.sha256()

        while True:
            data = self.file.read(256)
            if not data:
                break
            hash256.update(data)
        checksum = hash256.hexdigest()

        self.file.seek(original_position)
        return os.path.join(folder_name, checksum[:3], filename)


class CustomPageQuerySet(PageQuerySet):
    def about_spam(self):
        return self.filter(title__contains="spam")


CustomManager = PageManager.from_queryset(CustomPageQuerySet)


class CustomManagerPage(Page):
    objects = CustomManager()


class MyBasePage(Page):
    """
    A base Page model, used to set site-wide defaults and overrides.
    """

    objects = CustomManager()

    class Meta:
        abstract = True


class MyCustomPage(MyBasePage):
    pass


class ValidatedPage(Page):
    foo = models.CharField(max_length=255)

    base_form_class = ValidatedPageForm
    content_panels = Page.content_panels + [
        FieldPanel("foo"),
    ]


class DefaultRichTextFieldPage(Page):
    body = RichTextField()

    content_panels = [
        TitleFieldPanel("title", classname="title"),
        FieldPanel("body"),
    ]


class DefaultRichBlockFieldPage(Page):
    body = StreamField(
        [
            ("rich_text", RichTextBlock()),
        ],
    )

    content_panels = Page.content_panels + [FieldPanel("body")]


class CustomRichTextFieldPage(Page):
    body = RichTextField(editor="custom")

    content_panels = [
        TitleFieldPanel("title", classname="title"),
        FieldPanel("body"),
    ]


class CustomRichBlockFieldPage(Page):
    body = StreamField(
        [
            ("rich_text", RichTextBlock(editor="custom")),
        ],
    )

    content_panels = [
        TitleFieldPanel("title", classname="title"),
        FieldPanel("body"),
    ]


class RichTextFieldWithFeaturesPage(Page):
    body = RichTextField(features=["quotation", "embed", "made-up-feature"])

    content_panels = [
        TitleFieldPanel("title", classname="title"),
        FieldPanel("body"),
    ]


# a page that only contains RichTextField within an InlinePanel,
# to test that the inline child's form media gets pulled through
class SectionedRichTextPageSection(Orderable):
    page = ParentalKey(
        "tests.SectionedRichTextPage", related_name="sections", on_delete=models.CASCADE
    )
    body = RichTextField()

    panels = [FieldPanel("body")]


class SectionedRichTextPage(Page):
    content_panels = [
        TitleFieldPanel("title", classname="title"),
        InlinePanel("sections"),
    ]


class InlineStreamPageSection(Orderable):
    page = ParentalKey(
        "tests.InlineStreamPage", related_name="sections", on_delete=models.CASCADE
    )
    body = StreamField(
        [
            ("text", CharBlock()),
            ("rich_text", RichTextBlock()),
            ("image", ImageChooserBlock()),
        ],
    )
    panels = [FieldPanel("body")]


class InlineStreamPage(Page):
    content_panels = [
        TitleFieldPanel("title", classname="title"),
        InlinePanel("sections"),
    ]


class TableBlockStreamPage(Page):
    table = StreamField([("table", TableBlock())])

    content_panels = [FieldPanel("table")]


class UserProfile(models.Model):
    # Wagtail's schema must be able to coexist alongside a custom UserProfile model
    user = models.OneToOneField(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    favourite_colour = models.CharField(max_length=255)


class PanelSiteSettings(TestSiteSetting):
    panels = [FieldPanel("title")]


class PanelGenericSettings(TestGenericSetting):
    panels = [FieldPanel("title")]


class TabbedSiteSettings(TestSiteSetting):
    edit_handler = TabbedInterface(
        [
            ObjectList([FieldPanel("title")], heading="First tab"),
            ObjectList([FieldPanel("email")], heading="Second tab"),
        ]
    )


class TabbedGenericSettings(TestGenericSetting):
    edit_handler = TabbedInterface(
        [
            ObjectList([FieldPanel("title")], heading="First tab"),
            ObjectList([FieldPanel("email")], heading="Second tab"),
        ]
    )


class AlwaysShowInMenusPage(Page):
    show_in_menus_default = True


# test for AddField migrations on StreamFields using various default values
class AddedStreamFieldWithoutDefaultPage(Page):
    body = StreamField([("title", CharBlock())])


class AddedStreamFieldWithEmptyStringDefaultPage(Page):
    body = StreamField([("title", CharBlock())], default="")


class AddedStreamFieldWithEmptyListDefaultPage(Page):
    body = StreamField([("title", CharBlock())], default=[])


class SecretPage(Page):
    boring_data = models.TextField()
    secret_data = models.TextField()

    content_panels = Page.content_panels + [
        FieldPanel("boring_data"),
        FieldPanel("secret_data", permission="superuser"),
    ]


class SimpleParentPage(Page):
    subpage_types = ["tests.SimpleChildPage"]


class SimpleChildPage(Page):
    parent_page_types = ["tests.SimpleParentPage"]

    max_count_per_parent = 1


class PersonPage(Page):
    first_name = models.CharField(
        max_length=255,
        verbose_name="First Name",
    )
    last_name = models.CharField(
        max_length=255,
        verbose_name="Last Name",
    )

    content_panels = Page.content_panels + [
        MultiFieldPanel(
            [
                "first_name",
                "last_name",
            ],
            "Person",
        ),
        "addresses",
        "social_links",
    ]

    class Meta:
        verbose_name = "person"
        verbose_name_plural = "persons"


class Address(index.Indexed, ClusterableModel, Orderable):
    address = models.CharField(
        max_length=255,
        verbose_name="Address",
    )
    tags = ClusterTaggableManager(
        through="tests.AddressTag",
        blank=True,
    )
    person = ParentalKey(
        to="tests.PersonPage", related_name="addresses", verbose_name="Person"
    )

    panels = [
        FieldPanel("address"),
        FieldPanel("tags"),
    ]

    class Meta:
        verbose_name = "address"
        verbose_name_plural = "addresses"


class AddressTag(TaggedItemBase):
    content_object = ParentalKey(
        to="tests.Address", on_delete=models.CASCADE, related_name="tagged_items"
    )


class SocialLink(index.Indexed, ClusterableModel):
    url = models.URLField()
    kind = models.CharField(
        max_length=30,
        choices=[
            ("twitter", "Twitter"),
            ("facebook", "Facebook"),
        ],
    )
    person = ParentalKey(
        to="tests.PersonPage", related_name="social_links", verbose_name="Person"
    )

    panels = ["url", "kind"]

    class Meta:
        verbose_name = "social link"
        verbose_name_plural = "social links"


class RestaurantPage(Page):
    tags = ClusterTaggableManager(through="tests.TaggedRestaurant", blank=True)

    content_panels = Page.content_panels + [
        FieldPanel("tags"),
    ]


class RestaurantTag(TagBase):
    free_tagging = False

    class Meta:
        verbose_name = "tag"
        verbose_name_plural = "tags"


class TaggedRestaurant(ItemBase):
    tag = models.ForeignKey(
        RestaurantTag, related_name="tagged_restaurants", on_delete=models.CASCADE
    )
    content_object = ParentalKey(
        to="tests.RestaurantPage", on_delete=models.CASCADE, related_name="tagged_items"
    )


class SimpleTask(Task):
    pass


class UserApprovalTaskState(TaskState):
    pass


class UserApprovalTask(Task):
    """
    Based on https://docs.wagtail.org/en/stable/extending/custom_tasks.html.
    """

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.SET_NULL, null=True, blank=False
    )

    admin_form_fields = Task.admin_form_fields + ["user"]

    task_state_class = UserApprovalTaskState

    # prevent editing of `user` after the task is created
    # by default, this attribute contains the 'name' field to prevent tasks from being renamed
    admin_form_readonly_on_edit_fields = Task.admin_form_readonly_on_edit_fields + [
        "user"
    ]

    def user_can_access_editor(self, page, user):
        return user == self.user

    def page_locked_for_user(self, page, user):
        return user != self.user

    def get_actions(self, page, user):
        if user == self.user:
            return [
                ("approve", "Approve", False),
                ("approve", "Approve with style", True),
                ("reject", "Reject", False),
                ("cancel", "Cancel", False),
            ]
        else:
            return []

    def get_template_for_action(self, action):
        # https://github.com/wagtail/wagtail/issues/12222
        # This will be used for "Approve with style" which has the third value
        # (action_requires_additional_data_from_modal) set to True.
        if action == "approve":
            return "tests/workflows/approve_with_style.html"
        return super().get_template_for_action(action)

    def on_action(self, task_state, user, action_name, **kwargs):
        if action_name == "cancel":
            return task_state.workflow_state.cancel(user=user)
        else:
            return super().on_action(task_state, user, action_name, **kwargs)

    def get_task_states_user_can_moderate(self, user, **kwargs):
        if user == self.user:
            # get all task states linked to the (base class of) current task
            return TaskState.objects.filter(
                status=TaskState.STATUS_IN_PROGRESS, task=self.task_ptr
            )
        else:
            return TaskState.objects.none()

    @classmethod
    def get_description(cls):
        return "Only a specific user can approve this task"


# StreamField media definitions must not be evaluated at startup (e.g. during system checks) -
# these may fail if e.g. ManifestStaticFilesStorage is in use and collectstatic has not been run.
# Check this with a media definition that deliberately errors; if media handling is not set up
# correctly, then the mere presence of this model definition will cause startup to fail.
class DeadlyTextInput(forms.TextInput):
    @property
    def media(self):
        raise Exception("BOOM! Attempted to evaluate DeadlyTextInput.media")


class DeadlyCharBlock(FieldBlock):
    def __init__(self, *args, **kwargs):
        self.field = forms.CharField(widget=DeadlyTextInput())
        super().__init__(*args, **kwargs)


class DeadlyStreamPage(Page):
    body = StreamField(
        [
            ("title", DeadlyCharBlock()),
        ],
    )
    content_panels = Page.content_panels + [
        FieldPanel("body"),
    ]


# Check that get_image_model and get_document_model work at import time
# (so that it's possible to use them in foreign key definitions, for example)
ReimportedImageModel = get_image_model()
ReimportedDocumentModel = get_document_model()


# Custom document model with a custom tag field
class TaggedRestaurantDocument(ItemBase):
    tag = models.ForeignKey(
        RestaurantTag, related_name="tagged_documents", on_delete=models.CASCADE
    )
    content_object = models.ForeignKey(
        to="tests.CustomRestaurantDocument",
        on_delete=models.CASCADE,
        related_name="tagged_items",
    )


class CustomRestaurantDocument(AbstractDocument):
    tags = TaggableManager(
        help_text=None,
        blank=True,
        verbose_name="tags",
        through=TaggedRestaurantDocument,
    )
    admin_form_fields = Document.admin_form_fields


# Custom image model with a custom tag field
class TaggedRestaurantImage(ItemBase):
    tag = models.ForeignKey(
        RestaurantTag, related_name="tagged_images", on_delete=models.CASCADE
    )
    content_object = models.ForeignKey(
        to="tests.CustomRestaurantImage",
        on_delete=models.CASCADE,
        related_name="tagged_items",
    )


class CustomRestaurantImage(AbstractImage):
    tags = TaggableManager(
        help_text=None, blank=True, verbose_name="tags", through=TaggedRestaurantImage
    )
    admin_form_fields = Image.admin_form_fields


class ModelWithStringTypePrimaryKey(models.Model):
    """
    This model intentionally uses `CharField` as a primary key for testing purpose.
    """

    custom_id = models.CharField(max_length=255, primary_key=True)
    content = models.CharField(max_length=255)


class ModelWithNullableParentalKey(models.Model):
    """
    There's not really a valid use case for null parental keys, but their presence should not
    break things outright (e.g. when determining the object ID to store things under in the
    references index).
    """

    page = ParentalKey(Page, blank=True, null=True)
    content = RichTextField()


class GalleryPage(Page):
    content_panels = Page.content_panels + [
        MultipleChooserPanel(
            "gallery_images", heading="Gallery images", chooser_field_name="image"
        )
    ]


class GalleryPageImage(Orderable):
    page = ParentalKey(
        "tests.GalleryPage", related_name="gallery_images", on_delete=models.CASCADE
    )
    image = models.ForeignKey(
        "wagtailimages.Image",
        on_delete=models.CASCADE,
        related_name="+",
    )


class GenericSnippetNoIndexPage(GenericSnippetPage):
    wagtail_reference_index_ignore = True


class GenericSnippetNoFieldIndexPage(GenericSnippetPage):
    snippet_content_type_nonindexed = models.ForeignKey(
        ContentType, on_delete=models.SET_NULL, null=True, blank=True
    )
    snippet_content_type_nonindexed.wagtail_reference_index_ignore = True


def random_quotable_pk():
    quote_chrs = '":/_#?;@&=+$,"[]<>%\n\\'
    components = (quote_chrs, string.ascii_letters, string.digits)
    return "".join(random.choice(components[i % len(components)]) for i in range(10))


# Models to be registered with a ModelViewSet
class FeatureCompleteToy(index.Indexed, models.Model):
    strid = models.CharField(
        max_length=255,
        primary_key=True,
        default=random_quotable_pk,
    )
    name = models.CharField(max_length=255)
    release_date = models.DateField(default=datetime.date.today)

    search_fields = [
        index.SearchField("name"),
        index.AutocompleteField("name"),
        index.FilterField("name"),
        index.FilterField("release_date"),
    ]

    def is_cool(self):
        if self.name == self.name[::-1]:
            return True
        if (lowered := self.name.lower()) == lowered[::-1]:
            return None
        return False

    def __str__(self):
        return f"{self.name} ({self.release_date})"

    class Meta:
        permissions = [("can_set_release_date", "Can set release date")]


class PurgeRevisionsProtectedTestModel(models.Model):
    revision = models.OneToOneField(
        "wagtailcore.Revision", on_delete=models.PROTECT, related_name="+"
    )


class SearchTestModel(models.Model):
    title = models.CharField(max_length=255)
    body = models.TextField()
    panels = [
        FieldPanel("title"),
        FieldPanel("body"),
    ]

    def __str__(self):
        return self.title


class CustomPermissionTester(PagePermissionTester):
    def can_view_revisions(self):
        return False


class CustomPermissionPage(Page):
    def permissions_for_user(self, user):
        return CustomPermissionTester(user, self)


class CustomPermissionModel(models.Model):
    text = models.TextField(default="Tailwag")

    class Meta:
        verbose_name = "ADVANCED permission model"
        verbose_name_plural = "ADVANCED permission models"

        # Django's default_permissions are ("add", "change", "delete", "view").
        # Django will generate permissions for each of these actions with the
        # format f"{action}_{model_name}" and the label "Can {action} {verbose_name}".
        # This means if the action is "bulk_update", the codename will be
        # "bulk_update_custompermissionmodel" and the label will be
        # "Can bulk_update ADVANCED permission model".
        # See https://github.com/django/django/blob/stable/5.0.x/django/contrib/auth/management/__init__.py#L22-L35
        default_permissions = ("add", "change", "delete", "view", "bulk_update")

        # Permissions with completely custom codenames and labels
        permissions = [
            # Starts with can_ and "Can "
            ("can_start_trouble", "Can start trouble"),
            # Doesn't start with can_ and "Can "
            ("cause_chaos", "Cause chaos for advanced permission model"),
            # Starts with an action similar to a built-in permission "change_"
            ("change_text", "Change text"),
            # Without any _ and the label ends with the default verbose_name
            ("control", "Manage custom permission model"),
        ]


register_snippet(CustomPermissionModel)


class RequiredDatePage(Page):
    deadline = models.DateField()

    content_panels = [
        TitleFieldPanel("title", classname="title"),
        FieldPanel("deadline"),
    ]



from django.contrib.auth.models import Group, Permission
from django.test import TestCase, override_settings
from django.urls import reverse

from wagtail import hooks
from wagtail.admin.admin_url_finder import AdminURLFinder
from wagtail.admin.panels import FieldPanel, ObjectList, TabbedInterface
from wagtail.contrib.settings.registry import SettingMenuItem
from wagtail.contrib.settings.views import get_setting_edit_handler
from wagtail.models import GroupSitePermission, Page, Site
from wagtail.test.testapp.models import (
    FileSiteSetting,
    IconSiteSetting,
    PanelSiteSettings,
    TabbedSiteSettings,
    TestPermissionedSiteSetting,
    TestSiteSetting,
)
from wagtail.test.utils import WagtailTestUtils


class SiteSettingTestMixin:
    def login_only_admin(self):
        """Log in with a user that only has permission to access the admin"""
        user = self.create_user(username="test", password="password")
        user.user_permissions.add(
            Permission.objects.get_by_natural_key(
                codename="access_admin", app_label="wagtailadmin", model="admin"
            )
        )
        self.login(username="test", password="password")
        return user

    def grant_default_site_permission(self, user):
        """
        Grant the user permission to edit TestSiteSetting for the default site.
        This is used to test site-specific permissions.
        """
        site_owners = Group.objects.create(name="Site Owners")
        GroupSitePermission.objects.create(
            group=site_owners,
            site=Site.objects.get(is_default_site=True),
            permission=Permission.objects.get_by_natural_key(
                app_label="tests",
                model="testsitesetting",
                codename="change_testsitesetting",
            ),
        )
        user.groups.add(site_owners)


class TestSiteSettingMenu(SiteSettingTestMixin, WagtailTestUtils, TestCase):
    def test_menu_item_in_admin(self):
        self.login()
        response = self.client.get(reverse("wagtailadmin_home"))

        self.assertContains(response, "Test site setting")
        self.assertContains(
            response, reverse("wagtailsettings:edit", args=("tests", "testsitesetting"))
        )

        # test that custom icon is used
        self.assertContains(response, '"tag"')

    def test_menu_item_no_permissions(self):
        self.login_only_admin()
        response = self.client.get(reverse("wagtailadmin_home"))

        self.assertNotContains(response, "Test site setting")
        self.assertNotContains(
            response, reverse("wagtailsettings:edit", args=("tests", "testsitesetting"))
        )

    def test_menu_item_with_site_specific_permission(self):
        user = self.login_only_admin()
        self.grant_default_site_permission(user)

        response = self.client.get(reverse("wagtailadmin_home"))

        self.assertContains(response, "Test site setting")
        self.assertContains(
            response, reverse("wagtailsettings:edit", args=("tests", "testsitesetting"))
        )

    def test_menu_item_icon(self):
        menu_item = SettingMenuItem(IconSiteSetting, icon="tag", classname="test-class")
        self.assertEqual(menu_item.icon_name, "tag")
        self.assertEqual(menu_item.classname, "test-class")


class BaseTestSiteSettingView(WagtailTestUtils, TestCase):
    def get(self, site_pk=1, params={}, setting=TestSiteSetting):
        url = self.edit_url(setting=setting, site_pk=site_pk)
        return self.client.get(url, params)

    def post(self, site_pk=1, post_data={}, setting=TestSiteSetting):
        url = self.edit_url(setting=setting, site_pk=site_pk)
        return self.client.post(url, post_data)

    def edit_url(self, setting, site_pk=1):
        args = [setting._meta.app_label, setting._meta.model_name, site_pk]
        return reverse("wagtailsettings:edit", args=args)


class TestSiteSettingCreateView(SiteSettingTestMixin, BaseTestSiteSettingView):
    def setUp(self):
        self.user = self.login()

    def test_get_edit(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)

    def test_get_edit_without_permission(self):
        self.login_only_admin()
        response = self.get()
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse("wagtailadmin_home"))

    def test_get_edit_with_site_specific_permission(self):
        user = self.login_only_admin()
        self.grant_default_site_permission(user)
        response = self.get()
        self.assertEqual(response.status_code, 200)

    def test_get_edit_without_site_specific_permission(self):
        user = self.login_only_admin()
        self.grant_default_site_permission(user)
        other_site = Site.objects.create(
            hostname="example.com", root_page=Page.objects.get(pk=2)
        )
        response = self.get(site_pk=other_site.pk)
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse("wagtailadmin_home"))

    def test_edit_invalid(self):
        response = self.post(post_data={"foo": "bar"})
        self.assertContains(response, "The setting could not be saved due to errors.")
        self.assertContains(response, "error-message", count=2)
        self.assertContains(response, "This field is required", count=2)

    def test_edit(self):
        response = self.post(
            post_data={"title": "Edited site title", "email": "test@example.com"}
        )
        self.assertEqual(response.status_code, 302)

        default_site = Site.objects.get(is_default_site=True)
        setting = TestSiteSetting.objects.get(site=default_site)
        self.assertEqual(setting.title, "Edited site title")
        self.assertEqual(setting.email, "test@example.com")

        url_finder = AdminURLFinder(self.user)
        expected_url = "/admin/settings/tests/testsitesetting/%d/" % default_site.pk
        self.assertEqual(url_finder.get_edit_url(setting), expected_url)

    def test_edit_without_permission(self):
        self.login_only_admin()
        response = self.post(
            post_data={"title": "Edited site title", "email": "test@example.com"}
        )
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse("wagtailadmin_home"))
        self.assertFalse(TestSiteSetting.objects.exists())

    def test_edit_with_site_specific_permission(self):
        user = self.login_only_admin()
        self.grant_default_site_permission(user)

        response = self.post(
            post_data={"title": "Edited site title", "email": "test@example.com"}
        )
        self.assertEqual(response.status_code, 302)
        default_site = Site.objects.get(is_default_site=True)
        setting = TestSiteSetting.objects.get(site=default_site)
        self.assertEqual(setting.title, "Edited site title")
        self.assertEqual(setting.email, "test@example.com")

    def test_edit_without_site_specific_permission(self):
        user = self.login_only_admin()
        self.grant_default_site_permission(user)
        other_site = Site.objects.create(
            hostname="example.com", root_page=Page.objects.get(pk=2)
        )

        response = self.post(
            site_pk=other_site.pk,
            post_data={"title": "Edited site title", "email": "test@example.com"},
        )
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse("wagtailadmin_home"))
        self.assertFalse(TestSiteSetting.objects.exists())

    def test_file_upload_multipart(self):
        response = self.get(setting=FileSiteSetting)
        # Ensure the form supports file uploads
        self.assertContains(response, 'enctype="multipart/form-data"')

    def test_create_restricted_field_without_any_permission(self):
        # User has no permissions over the setting model, only access to the admin
        self.user.is_superuser = False
        self.user.save()
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            ),
        )

        self.assertFalse(TestPermissionedSiteSetting.objects.exists())
        # GET should redirect away with permission denied
        response = self.get(setting=TestPermissionedSiteSetting)
        self.assertRedirects(response, status_code=302, expected_url="/admin/")

        # the GET might create a setting object, depending on when the permission check is done,
        # so remove any created objects prior to testing the POST

        # POST should redirect away with permission denied
        response = self.post(
            post_data={"sensitive_email": "test@example.com", "title": "test"},
            setting=TestPermissionedSiteSetting,
        )
        self.assertRedirects(response, status_code=302, expected_url="/admin/")

        # The retrieved setting should contain none of the submitted data
        settings = TestPermissionedSiteSetting.for_site(Site.objects.get(pk=1))
        self.assertEqual(settings.title, "")
        self.assertEqual(settings.sensitive_email, "")

    def test_create_restricted_field_without_field_permission(self):
        # User has edit permission over the setting model, but not the sensitive_email field
        self.user.is_superuser = False
        self.user.save()
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            ),
            Permission.objects.get(
                content_type__app_label="tests",
                codename="change_testpermissionedsitesetting",
            ),
        )

        self.assertFalse(TestPermissionedSiteSetting.objects.exists())
        # GET should provide a form with title but not sensitive_email
        response = self.get(setting=TestPermissionedSiteSetting)
        self.assertEqual(response.status_code, 200)
        self.assertIn("title", list(response.context["form"].fields))
        self.assertNotIn("sensitive_email", list(response.context["form"].fields))

        # the GET creates a setting object, so remove any created objects prior to testing the POST
        TestPermissionedSiteSetting.objects.all().delete()

        # POST should allow the title to be set, but not the sensitive_email
        response = self.post(
            post_data={"sensitive_email": "test@example.com", "title": "test"},
            setting=TestPermissionedSiteSetting,
        )
        self.assertEqual(response.status_code, 302)

        settings = TestPermissionedSiteSetting.objects.get()
        self.assertEqual(settings.title, "test")
        self.assertEqual(settings.sensitive_email, "")

    def test_create_restricted_field(self):
        # User has edit permission over the setting model, including the sensitive_email field
        self.user.is_superuser = False
        self.user.save()
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            ),
            Permission.objects.get(
                content_type__app_label="tests",
                codename="change_testpermissionedsitesetting",
            ),
            Permission.objects.get(codename="can_edit_sensitive_email_site_setting"),
        )
        self.assertFalse(TestPermissionedSiteSetting.objects.exists())
        # GET should provide a form with title and sensitive_email
        response = self.get(setting=TestPermissionedSiteSetting)
        self.assertEqual(response.status_code, 200)
        self.assertIn("title", list(response.context["form"].fields))
        self.assertIn("sensitive_email", list(response.context["form"].fields))

        # the GET creates a setting object, so remove any created objects prior to testing the POST
        TestPermissionedSiteSetting.objects.all().delete()

        # POST should allow both title and sensitive_email to be set
        response = self.post(
            post_data={"sensitive_email": "test@example.com", "title": "test"},
            setting=TestPermissionedSiteSetting,
        )
        self.assertEqual(response.status_code, 302)

        settings = TestPermissionedSiteSetting.objects.get()
        self.assertEqual(settings.title, "test")
        self.assertEqual(settings.sensitive_email, "test@example.com")


class TestSiteSettingEditView(SiteSettingTestMixin, BaseTestSiteSettingView):
    def setUp(self):
        self.default_site = Site.objects.get(is_default_site=True)

        self.test_setting = TestSiteSetting()
        self.test_setting.title = "Site title"
        self.test_setting.email = "initial@example.com"
        self.test_setting.site = self.default_site
        self.test_setting.save()

        self.user = self.login()

    def test_get_edit(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)

    def test_get_edit_without_permission(self):
        self.login_only_admin()
        response = self.get()
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse("wagtailadmin_home"))

    def test_get_edit_with_site_specific_permission(self):
        user = self.login_only_admin()
        self.grant_default_site_permission(user)
        response = self.get()
        self.assertEqual(response.status_code, 200)

    def test_get_edit_without_site_specific_permission(self):
        user = self.login_only_admin()
        self.grant_default_site_permission(user)
        other_site = Site.objects.create(
            hostname="example.com", root_page=Page.objects.get(pk=2)
        )
        response = self.get(site_pk=other_site.pk)
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse("wagtailadmin_home"))

    def test_non_existent_model(self):
        response = self.client.get(
            reverse("wagtailsettings:edit", args=["test", "foo", 1])
        )
        self.assertEqual(response.status_code, 404)

    def test_register_with_icon(self):
        edit_url = reverse("wagtailsettings:edit", args=("tests", "IconGenericSetting"))
        edit_response = self.client.get(edit_url, follow=True)
        soup = self.get_soup(edit_response.content)
        self.assertIsNotNone(soup.select_one("h2 svg use[href='#icon-tag']"))

    def test_edit_invalid(self):
        response = self.post(post_data={"foo": "bar"})
        self.assertContains(response, "The setting could not be saved due to errors.")
        self.assertContains(response, "error-message", count=2)
        self.assertContains(response, "This field is required", count=2)

    def test_edit(self):
        response = self.post(
            post_data={"title": "Edited site title", "email": "test@example.com"}
        )
        self.assertEqual(response.status_code, 302)

        default_site = Site.objects.get(is_default_site=True)
        setting = TestSiteSetting.objects.get(site=default_site)
        self.assertEqual(setting.title, "Edited site title")
        self.assertEqual(setting.email, "test@example.com")

    def test_edit_without_permission(self):
        self.login_only_admin()
        response = self.post(
            post_data={"title": "Edited site title", "email": "test@example.com"}
        )
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse("wagtailadmin_home"))
        self.test_setting.refresh_from_db()
        self.assertEqual(self.test_setting.title, "Site title")
        self.assertEqual(self.test_setting.email, "initial@example.com")

    def test_edit_with_site_specific_permission(self):
        user = self.login_only_admin()
        self.grant_default_site_permission(user)

        response = self.post(
            post_data={"title": "Edited site title", "email": "test@example.com"}
        )
        self.assertEqual(response.status_code, 302)
        self.test_setting.refresh_from_db()
        self.assertEqual(self.test_setting.title, "Edited site title")
        self.assertEqual(self.test_setting.email, "test@example.com")

    def test_edit_without_site_specific_permission(self):
        user = self.login_only_admin()
        self.grant_default_site_permission(user)
        other_site = Site.objects.create(
            hostname="example.com", root_page=Page.objects.get(pk=2)
        )

        response = self.post(
            site_pk=other_site.pk,
            post_data={"title": "Edited site title", "email": "test@example.com"},
        )
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse("wagtailadmin_home"))
        self.test_setting.refresh_from_db()
        self.assertEqual(self.test_setting.title, "Site title")
        self.assertEqual(self.test_setting.email, "initial@example.com")

    def test_get_redirect_to_relevant_instance(self):
        url = reverse("wagtailsettings:edit", args=("tests", "testsitesetting"))
        default_site = Site.objects.get(is_default_site=True)

        response = self.client.get(url)
        self.assertRedirects(
            response, status_code=302, expected_url=f"{url}{default_site.pk}/"
        )

    def test_get_redirect_to_relevant_instance_invalid(self):
        Site.objects.all().delete()
        url = reverse("wagtailsettings:edit", args=("tests", "testsitesetting"))
        response = self.client.get(url)
        self.assertRedirects(response, status_code=302, expected_url="/admin/")

    def test_get_redirect_to_relevant_instance_with_specific_site_permission(self):
        user = self.login_only_admin()
        other_site = Site.objects.create(
            hostname="example.com", root_page=Page.objects.get(pk=2)
        )

        # grant this user permission to edit other_site only
        site_owners = Group.objects.create(name="Site Owners")
        GroupSitePermission.objects.create(
            group=site_owners,
            site=other_site,
            permission=Permission.objects.get_by_natural_key(
                app_label="tests",
                model="testsitesetting",
                codename="change_testsitesetting",
            ),
        )
        user.groups.add(site_owners)

        url = reverse("wagtailsettings:edit", args=("tests", "testsitesetting"))

        response = self.client.get(url)
        self.assertRedirects(
            response, status_code=302, expected_url=f"{url}{other_site.pk}/"
        )

    def test_edit_restricted_field(self):
        # User has edit permission over the setting model, including the sensitive_email field
        test_setting = TestPermissionedSiteSetting()
        test_setting.title = "Old title"
        test_setting.sensitive_email = "test@example.com"
        test_setting.site = self.default_site
        test_setting.save()
        self.user.is_superuser = False
        self.user.save()
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            ),
            Permission.objects.get(
                content_type__app_label="tests",
                codename="change_testpermissionedsitesetting",
            ),
            Permission.objects.get(codename="can_edit_sensitive_email_site_setting"),
        )

        # GET should provide a form with title and sensitive_email
        response = self.get(setting=TestPermissionedSiteSetting)
        self.assertEqual(response.status_code, 200)
        self.assertIn("title", list(response.context["form"].fields))
        self.assertIn("sensitive_email", list(response.context["form"].fields))

        # POST should allow both title and sensitive_email to be set
        response = self.post(
            setting=TestPermissionedSiteSetting,
            post_data={
                "sensitive_email": "test-updated@example.com",
                "title": "New title",
            },
        )
        self.assertEqual(response.status_code, 302)

        test_setting.refresh_from_db()
        self.assertEqual(test_setting.sensitive_email, "test-updated@example.com")
        self.assertEqual(test_setting.title, "New title")

    def test_edit_restricted_field_without_field_permission(self):
        # User has edit permission over the setting model, but not the sensitive_email field
        test_setting = TestPermissionedSiteSetting()
        test_setting.title = "Old title"
        test_setting.sensitive_email = "test@example.com"
        test_setting.site = self.default_site
        test_setting.save()
        self.user.is_superuser = False
        self.user.save()
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            ),
            Permission.objects.get(
                content_type__app_label="tests",
                codename="change_testpermissionedsitesetting",
            ),
        )

        # GET should provide a form with title but not sensitive_email
        response = self.get(setting=TestPermissionedSiteSetting)
        self.assertEqual(response.status_code, 200)
        self.assertIn("title", list(response.context["form"].fields))
        self.assertNotIn("sensitive_email", list(response.context["form"].fields))

        # POST should allow the title to be set, but not the sensitive_email
        response = self.post(
            setting=TestPermissionedSiteSetting,
            post_data={
                "sensitive_email": "test-updated@example.com",
                "title": "New title",
            },
        )
        self.assertEqual(response.status_code, 302)

        test_setting.refresh_from_db()
        self.assertEqual(test_setting.sensitive_email, "test@example.com")
        self.assertEqual(test_setting.title, "New title")

    def test_edit_restricted_field_without_any_permission(self):
        # User has no permissions over the setting model, only access to the admin
        test_setting = TestPermissionedSiteSetting()
        test_setting.title = "Old title"
        test_setting.sensitive_email = "test@example.com"
        test_setting.site = self.default_site
        test_setting.save()
        self.user.is_superuser = False
        self.user.save()
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            ),
        )

        # GET should redirect away with permission denied
        response = self.get(setting=TestPermissionedSiteSetting)
        self.assertRedirects(response, status_code=302, expected_url="/admin/")

        # POST should redirect away with permission denied
        response = self.post(
            setting=TestPermissionedSiteSetting,
            post_data={
                "sensitive_email": "test-updated@example.com",
                "title": "New title",
            },
        )
        self.assertRedirects(response, status_code=302, expected_url="/admin/")

        # The retrieved setting should be unchanged
        test_setting.refresh_from_db()
        self.assertEqual(test_setting.sensitive_email, "test@example.com")
        self.assertEqual(test_setting.title, "Old title")


@override_settings(
    ALLOWED_HOSTS=["testserver", "example.com", "noneoftheabove.example.com"]
)
class TestMultiSite(SiteSettingTestMixin, BaseTestSiteSettingView):
    def setUp(self):
        self.default_site = Site.objects.get(is_default_site=True)
        self.other_site = Site.objects.create(
            hostname="example.com", root_page=Page.objects.get(pk=2)
        )
        self.login()

    def test_redirect_to_default(self):
        """
        Should redirect to the setting for the default site.
        """
        start_url = reverse("wagtailsettings:edit", args=["tests", "testsitesetting"])
        dest_url = reverse(
            "wagtailsettings:edit",
            args=["tests", "testsitesetting", self.default_site.pk],
        )
        response = self.client.get(start_url, follow=True)
        self.assertRedirects(
            response, dest_url, status_code=302, fetch_redirect_response=False
        )

    def test_redirect_to_current(self):
        """
        Should redirect to the setting for the current site taken from the URL,
        by default
        """
        start_url = reverse("wagtailsettings:edit", args=["tests", "testsitesetting"])
        dest_url = reverse(
            "wagtailsettings:edit",
            args=["tests", "testsitesetting", self.other_site.pk],
        )
        response = self.client.get(
            start_url, follow=True, HTTP_HOST=self.other_site.hostname
        )
        self.assertRedirects(
            response, dest_url, status_code=302, fetch_redirect_response=False
        )

    def test_with_no_current_site(self):
        """
        Redirection should not break if the current request does not correspond to a site
        """
        self.default_site.is_default_site = False
        self.default_site.save()

        start_url = reverse("wagtailsettings:edit", args=["tests", "testsitesetting"])
        response = self.client.get(
            start_url, follow=True, HTTP_HOST="noneoftheabove.example.com"
        )
        self.assertEqual(302, response.redirect_chain[0][1])

    def test_switcher(self):
        """Check that the switcher form exists in the page"""
        response = self.get()
        self.assertEqual(response.status_code, 200)
        soup = self.get_soup(response.content)
        options = soup.select("form#settings-site-switch option")
        self.assertEqual(len(options), 2)  # other site + default site
        self.assertEqual("example.com", options[0].text.strip())
        self.assertEqual("localhost [default]", options[1].text.strip())

    def test_no_switcher_when_only_permission_for_one_site(self):
        """
        The switcher should not be displayed if the user only has permission to edit one site,
        but a static label should be shown instead
        """
        user = self.login_only_admin()
        self.grant_default_site_permission(user)
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertNotContains(response, 'id="settings-site-switch"')
        self.assertContains(response, "Site: localhost [default]", html=True)

    def test_no_switcher_when_only_one_site_exists(self):
        """
        No switcher or label should be shown if only one site exists
        """
        user = self.login_only_admin()
        self.grant_default_site_permission(user)
        self.other_site.delete()  # Remove the other site

        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertNotContains(response, 'id="settings-site-switch"')
        self.assertNotContains(response, "Site: localhost [default]", html=True)

    def test_switcher_when_permission_for_individual_sites(self):
        """The switcher should only show the sites the user has permission to edit"""
        user = self.login_only_admin()
        self.grant_default_site_permission(user)
        third_site = Site.objects.create(
            hostname="third.example.com", root_page=Page.objects.get(pk=2)
        )
        GroupSitePermission.objects.create(
            group=Group.objects.get(name="Site Owners"),
            site=third_site,
            permission=Permission.objects.get_by_natural_key(
                app_label="tests",
                model="testsitesetting",
                codename="change_testsitesetting",
            ),
        )

        response = self.get()
        self.assertEqual(response.status_code, 200)
        soup = self.get_soup(response.content)
        options = soup.select("form#settings-site-switch option")
        self.assertEqual(len(options), 2)  # default site + third site
        self.assertEqual("localhost [default]", options[0].text.strip())
        self.assertEqual("third.example.com", options[1].text.strip())

    def test_unknown_site(self):
        """Check that unknown sites throw a 404"""
        response = self.get(site_pk=3)
        self.assertEqual(response.status_code, 404)

    def test_edit(self):
        """
        Check that editing settings in multi-site mode edits the correct
        setting, and leaves the other ones alone
        """
        TestSiteSetting.objects.create(
            title="default", email="default@example.com", site=self.default_site
        )
        TestSiteSetting.objects.create(
            title="other", email="other@example.com", site=self.other_site
        )
        response = self.post(
            site_pk=self.other_site.pk,
            post_data={"title": "other-new", "email": "other-other@example.com"},
        )
        self.assertEqual(response.status_code, 302)

        # Check that the correct setting was updated
        other_setting = TestSiteSetting.for_site(self.other_site)
        self.assertEqual(other_setting.title, "other-new")
        self.assertEqual(other_setting.email, "other-other@example.com")

        # Check that the other setting was not updated
        default_setting = TestSiteSetting.for_site(self.default_site)
        self.assertEqual(default_setting.title, "default")
        self.assertEqual(default_setting.email, "default@example.com")


class TestAdminPermission(WagtailTestUtils, TestCase):
    def test_registered_permission(self):
        permission = Permission.objects.get_by_natural_key(
            app_label="tests",
            model="testsitesetting",
            codename="change_testsitesetting",
        )
        for fn in hooks.get_hooks("register_permissions"):
            if permission in fn():
                break
        else:
            self.fail("Change permission for tests.TestSiteSetting not registered")


class TestEditHandlers(TestCase):
    def setUp(self):
        get_setting_edit_handler.cache_clear()

    def test_default_model_introspection(self):
        handler = get_setting_edit_handler(TestSiteSetting)
        self.assertIsInstance(handler, ObjectList)
        self.assertEqual(len(handler.children), 2)
        first = handler.children[0]
        self.assertIsInstance(first, FieldPanel)
        self.assertEqual(first.field_name, "title")
        second = handler.children[1]
        self.assertIsInstance(second, FieldPanel)
        self.assertEqual(second.field_name, "email")

    def test_with_custom_panels(self):
        handler = get_setting_edit_handler(PanelSiteSettings)
        self.assertIsInstance(handler, ObjectList)
        self.assertEqual(len(handler.children), 1)
        first = handler.children[0]
        self.assertIsInstance(first, FieldPanel)
        self.assertEqual(first.field_name, "title")

    def test_with_custom_edit_handler(self):
        handler = get_setting_edit_handler(TabbedSiteSettings)
        self.assertIsInstance(handler, TabbedInterface)
        self.assertEqual(len(handler.children), 2)


class TestPermissionConfiguration(WagtailTestUtils, TestCase):
    def setUp(self):
        self.login()

        self.group = Group.objects.get(name="Editors")
        self.default_site = Site.objects.get(is_default_site=True)
        self.other_site = Site.objects.create(
            hostname="example.com", root_page=Page.objects.get(pk=2)
        )
        self.permission = Permission.objects.get_by_natural_key(
            app_label="tests",
            model="testsitesetting",
            codename="change_testsitesetting",
        )
        GroupSitePermission.objects.create(
            group=self.group,
            site=self.default_site,
            permission=self.permission,
        )

    def test_get_permissions(self):
        response = self.client.get(
            reverse("wagtailusers_groups:edit", args=(self.group.id,)),
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Test site setting permissions")
        soup = self.get_soup(response.content)
        default_site_checkbox = soup.select_one(
            f"[name='tests_testsitesetting_site_permissions-sites'][value='{self.default_site.pk}']"
        )
        self.assertTrue(default_site_checkbox.has_attr("checked"))
        other_site_checkbox = soup.select_one(
            f"[name='tests_testsitesetting_site_permissions-sites'][value='{self.other_site.pk}']"
        )
        self.assertFalse(other_site_checkbox.has_attr("checked"))

    def test_set_permissions(self):
        response = self.client.post(
            reverse("wagtailusers_groups:edit", args=(self.group.id,)),
            {
                "name": "test group",
                "permissions": [],
                "page_permissions-TOTAL_FORMS": ["0"],
                "page_permissions-MAX_NUM_FORMS": ["1000"],
                "page_permissions-INITIAL_FORMS": ["0"],
                "document_permissions-TOTAL_FORMS": ["0"],
                "document_permissions-MAX_NUM_FORMS": ["1000"],
                "document_permissions-INITIAL_FORMS": ["0"],
                "image_permissions-TOTAL_FORMS": ["0"],
                "image_permissions-MAX_NUM_FORMS": ["1000"],
                "image_permissions-INITIAL_FORMS": ["0"],
                "collection_permissions-TOTAL_FORMS": ["0"],
                "collection_permissions-MAX_NUM_FORMS": ["1000"],
                "collection_permissions-INITIAL_FORMS": ["0"],
                "tests_testsitesetting_site_permissions-sites": [
                    str(self.other_site.pk),
                ],
            },
        )
        self.assertRedirects(response, reverse("wagtailusers_groups:index"))

        self.assertTrue(
            GroupSitePermission.objects.filter(
                group=self.group,
                site=self.other_site,
                permission=self.permission,
            ).exists()
        )
        self.assertFalse(
            GroupSitePermission.objects.filter(
                group=self.group,
                site=self.default_site,
                permission=self.permission,
            ).exists()
        )



import datetime
import json

from django.core.exceptions import ValidationError
from django.core.serializers.json import DjangoJSONEncoder
from django.core.validators import BaseValidator, MaxLengthValidator
from django.db import models
from django.utils.encoding import force_str
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from wagtail.blocks import Block, BlockField, StreamBlock, StreamValue
from wagtail.blocks.definition_lookup import (
    BlockDefinitionLookup,
    BlockDefinitionLookupBuilder,
)
from wagtail.rich_text import (
    RichTextMaxLengthValidator,
    extract_references_from_rich_text,
    get_text_for_indexing,
)


class NoFutureDateValidator(BaseValidator):
    """
    A validator that prevents future dates from being entered.

    Useful for fields that should be in the past or present
    but never in the future.
    """

    message = _("Date cannot be in the future.")
    code = "future_date"

    def __init__(self, message=None):
        super().__init__(limit_value=None, message=message)

    def __call__(self, value):
        if value and value > datetime.date.today():
            raise ValidationError(self.message, code=self.code)


class RichTextField(models.TextField):
    def __init__(self, *args, **kwargs):
        # 'editor' and 'features' are popped before super().__init__ has chance to capture them
        # for use in deconstruct(). This is intentional - they would not be useful in migrations
        # and retrospectively adding them would generate unwanted migration noise
        self.editor = kwargs.pop("editor", "default")
        self.features = kwargs.pop("features", None)

        super().__init__(*args, **kwargs)

    def clone(self):
        name, path, args, kwargs = self.deconstruct()
        # add back the 'features' and 'editor' kwargs that were not preserved by deconstruct()
        kwargs["features"] = self.features
        kwargs["editor"] = self.editor
        return self.__class__(*args, **kwargs)

    def formfield(self, **kwargs):
        from wagtail.admin.rich_text import get_rich_text_editor_widget

        defaults = {
            "widget": get_rich_text_editor_widget(self.editor, features=self.features)
        }
        defaults.update(kwargs)
        field = super().formfield(**defaults)

        # replace any MaxLengthValidators with RichTextMaxLengthValidators to ignore tags
        for i, validator in enumerate(field.validators):
            if isinstance(validator, MaxLengthValidator):
                field.validators[i] = RichTextMaxLengthValidator(
                    validator.limit_value, message=validator.message
                )

        return field

    def get_searchable_content(self, value):
        # Strip HTML tags to prevent search backend from indexing them
        source = force_str(value)
        return [get_text_for_indexing(source)]

    def extract_references(self, value):
        yield from extract_references_from_rich_text(force_str(value))


# https://github.com/django/django/blob/64200c14e0072ba0ffef86da46b2ea82fd1e019a/django/db/models/fields/subclassing.py#L31-L44
class Creator:
    """
    A placeholder class that provides a way to set the attribute on the model.
    """

    def __init__(self, field):
        self.field = field

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        field_name = self.field.name

        if field_name not in obj.__dict__:
            # Field is deferred. Fetch it from db.
            obj.refresh_from_db(fields=[field_name])
        return obj.__dict__[field_name]

    def __set__(self, obj, value):
        obj.__dict__[self.field.name] = self.field.to_python(value)


class StreamField(models.Field):
    def __init__(self, block_types, use_json_field=True, block_lookup=None, **kwargs):
        """
        Construct a StreamField.

        :param block_types: Either a list of block types that are allowed in this StreamField
            (as a list of tuples of block name and block instance) or a StreamBlock to use as
            the top level block (as a block instance or class).
        :param use_json_field: Ignored, but retained for compatibility with historical migrations.
        :param block_lookup: Used in migrations to provide a more compact block definition -
            see ``wagtail.blocks.definition_lookup.BlockDefinitionLookup``. If passed, ``block_types``
            can contain integer indexes into this lookup table, in place of actual block instances.
        """

        # extract kwargs that are to be passed on to the block, not handled by super
        self.block_opts = {}
        for arg in ["min_num", "max_num", "block_counts", "collapsed"]:
            if arg in kwargs:
                self.block_opts[arg] = kwargs.pop(arg)

        # for a top-level block, the 'blank' kwarg (defaulting to False) always overrides the
        # block's own 'required' meta attribute, even if not passed explicitly; this ensures
        # that the field and block have consistent definitions
        self.block_opts["required"] = not kwargs.get("blank", False)

        # Store the `block_types` and `block_lookup` arguments to be handled in the `stream_block`
        # property
        self.block_types_arg = block_types
        self.block_lookup = block_lookup

        super().__init__(**kwargs)

    @cached_property
    def stream_block(self):
        has_block_lookup = self.block_lookup is not None
        if has_block_lookup:
            lookup = BlockDefinitionLookup(self.block_lookup)

        if isinstance(self.block_types_arg, Block):
            # use the passed block as the top-level block
            block = self.block_types_arg
        elif isinstance(self.block_types_arg, int) and has_block_lookup:
            # retrieve block from lookup table to use as the top-level block
            block = lookup.get_block(self.block_types_arg)
        elif isinstance(self.block_types_arg, type):
            # block passed as a class - instantiate it
            block = self.block_types_arg()
        else:
            # construct a top-level StreamBlock from the list of block types.
            # If an integer is found in place of a block instance, and block_lookup is
            # provided, it will be replaced with the corresponding block definition.
            child_blocks = []

            for name, child_block in self.block_types_arg:
                if isinstance(child_block, int) and has_block_lookup:
                    child_blocks.append((name, lookup.get_block(child_block)))
                else:
                    child_blocks.append((name, child_block))

            block = StreamBlock(child_blocks)

        if not isinstance(block, StreamBlock):
            raise TypeError(
                f"The top-level block must be a StreamBlock (got {type(block).__name__}). "
                "Either pass a StreamBlock instance/class, or a list of block definitions "
                "as (name, block) tuples."
            )

        block.set_meta_options(self.block_opts)
        return block

    @property
    def json_field(self):
        return models.JSONField(encoder=DjangoJSONEncoder)

    def get_internal_type(self):
        return "JSONField"

    def get_lookup(self, lookup_name):
        return self.json_field.get_lookup(lookup_name)

    def get_transform(self, lookup_name):
        return self.json_field.get_transform(lookup_name)

    def deconstruct(self):
        name, path, _, kwargs = super().deconstruct()
        lookup = BlockDefinitionLookupBuilder()
        block_types = [
            (name, lookup.add_block(block))
            for name, block in self.stream_block.child_blocks.items()
        ]
        args = [block_types]
        kwargs["block_lookup"] = lookup.get_lookup_as_dict()
        return name, path, args, kwargs

    def to_python(self, value):
        result = self.stream_block.to_python(value)

        # The top-level StreamValue is passed a reference to the StreamField, to support
        # pickling. This is necessary because unpickling needs access to the StreamBlock
        # definition, which cannot itself be pickled; instead we store a pointer to the
        # field within the model, which gives us a path to retrieve the StreamBlock definition.
        result._stream_field = self
        return result

    def get_prep_value(self, value):
        value = super().get_prep_value(value)
        if (
            isinstance(value, StreamValue)
            and not (value)
            and value.raw_text is not None
        ):
            # An empty StreamValue with a nonempty raw_text attribute should have that
            # raw_text attribute written back to the db. (This is probably only useful
            # for reverse migrations that convert StreamField data back into plain text
            # fields.)
            return value.raw_text
        elif isinstance(value, StreamValue):
            # StreamValue instances must be prepared first.
            return self.stream_block.get_prep_value(value)
        else:
            # If the value is not a StreamValue, it's likely the field is being
            # used in a non-Wagtail context, e.g. in queries with JSONField features.
            return super().get_prep_value(value)

    def get_db_prep_value(self, value, connection, prepared=False):
        # Use JSONField's get_db_prep_value method to handle the serialization,
        # which may differ between database backends. However, use our own
        # get_prep_value method to ensure that StreamValue instances are prepared
        # before being passed to JSONField.
        if not prepared:
            value = self.get_prep_value(value)
        return self.json_field.get_db_prep_value(
            value, connection=connection, prepared=True
        )

    def from_db_value(self, value, expression, connection):
        # Historically, StreamField's deserialization used to be handled by
        # to_python, which in turn handled by BaseStreamBlock.to_python. This was
        # always the case even before and after the use of the JSON data type.

        # However, now that we can be confident all StreamField data has been
        # migrated to use JSON in the database, we can reuse any special handling
        # that JSONField.from_db_value provides, e.g. for handling KeyTransforms
        # on SQLite.

        # This means we are passing a deserialized value to StreamBlock.to_python,
        # which is a change from the previous behaviour. However, this is fine
        # because to_python can handle both serialized and deserialized values.
        value = self.json_field.from_db_value(value, expression, connection)
        return self.to_python(value)

    def formfield(self, **kwargs):
        """
        Override formfield to use a plain forms.Field so that we do no transformation on the value
        (as distinct from the usual fallback of forms.CharField, which transforms it into a string).
        """
        defaults = {"form_class": BlockField, "block": self.stream_block}
        defaults.update(kwargs)
        return super().formfield(**defaults)

    def get_default(self):
        return self.stream_block.normalize(super().get_default())

    def value_to_string(self, obj):
        # This method is used for serialization using django.core.serializers,
        # which is used by dumpdata and loaddata for serializing model objects.
        # Unlike other fields, JSONField only uses value_from_object without
        # doing the actual serialization, so that it doesn't end up being
        # double-serialized when the model object is serialized.

        # Unfortunately, this is also used by django-modelcluster, which is used
        # to serialize model objects to be stored in revisions. When we migrated
        # StreamField to use the JSON data type, we did not change this method's
        # behaviour, i.e. it still returns a JSON-shaped string, to ensure that
        # revisions are still saved in the same format as before – even if it
        # means StreamField inside the revision data becomes double-serialized.

        # Now that we change get_prep_value to not do the serialization in favor
        # of get_db_prep_value, we need to add the serialization here too.
        value = self.value_from_object(obj)
        return json.dumps(self.get_prep_value(value), cls=self.json_field.encoder)

    def get_searchable_content(self, value):
        return self.stream_block.get_searchable_content(value)

    def extract_references(self, value):
        yield from self.stream_block.extract_references(value)

    def get_block_by_content_path(self, value, path_elements):
        """
        Given a list of elements from a content path, retrieve the block at that path
        as a BoundBlock object, or None if the path does not correspond to a valid block.
        """
        return self.stream_block.get_block_by_content_path(value, path_elements)

    def check(self, **kwargs):
        errors = super().check(**kwargs)
        errors.extend(self.stream_block.check(field=self, **kwargs))
        return errors

    def contribute_to_class(self, cls, name, **kwargs):
        super().contribute_to_class(cls, name, **kwargs)

        # Add Creator descriptor to allow the field to be set from a list or a
        # JSON string.
        setattr(cls, self.name, Creator(self))
